from .digflow_identifier import *
from .digiflow_io import *
from .digiflow_metadata import *
from .digiflow_generate import *
from .digiflow_export import *
from .digiflow_validate import (
    DDB_IGNORE_RULES_BASIC,
    DDB_IGNORE_RULES_MVW,
    DDB_IGNORE_RULES_NEWSPAPERS,
    DDB_MEDIA_SCH,
    DDB_NEWSP_SCH,
	DDB_ERROR,
	DDB_OTHER,
    CLI_DIR,
    PATH_CLI,
    SCH_DIR,
    PATH_MEDIA_SCH,
    PATH_NEWSP_SCH,
    DIGIS_MULTIVOLUME,
    DIGIS_NEWSPAPER,
    REPORT_FILE_SCHEMATRON,
	DigiflowDDBException,
    ddb_validation,
    ddb_validation_sch,
)



# -*- coding: utf-8 -*-
"""Digiflow validation API"""

import os
import shutil
from pathlib import (
    Path
)
from unittest import (
    mock
)

import pytest

from digiflow.digiflow_validate import (
    DDB_IGNORE_RULES_MVW,
    REPORT_FILE_SCHEMATRON,
    DDB_ERROR,
    DDB_OTHER,
    DigiflowDDBException,
    ddb_validation,
    ddb_validation_sch,
)

from .conftest import (
    TEST_RES
)

# switch off saxon py if not available
SAXON_PY_ENABLED = True
try:
    import saxonche
except ModuleNotFoundError: 
    SAXON_PY_ENABLED = False

ROOT = Path(__file__).parents[1]
PROJECT_ROOT_DIR = Path(__file__).resolve().parents[1]
SCHEMATRON_DIR = PROJECT_ROOT_DIR / 'src' / 'digiflow' / 'schematron'
SCHEMATRON_BIN = str(SCHEMATRON_DIR / 'schxslt-cli.jar')


@pytest.fixture(name="share_it_monography")
def _fixture_share_it_monography(tmp_path):
    """Provide test fixture"""

    testroot = tmp_path / 'VALIDATIONTEST'
    testroot.mkdir()
    file_name = '1981185920_44046.xml'
    mets_source = Path(TEST_RES) / 'ocr' / file_name
    mets_target = Path(str(testroot), file_name)
    shutil.copy(str(mets_source), str(mets_target))
    return str(mets_target)


def _place_file(path_dir: Path, mets_file_name: str):
    """Little Helper to push test data"""
    _src_name = f'sch_result_{mets_file_name}'
    _src_path = Path(TEST_RES) / _src_name
    _dst_path = path_dir / REPORT_FILE_SCHEMATRON
    shutil.copy(str(_src_path), str(_dst_path))
    return str(_dst_path)


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
@mock.patch('digiflow.digiflow_validate._forward_sch_cli')
def test_ddb_schematron_opendata_44046_default(sch_mock, share_it_monography):
    """Schematron validation with common monography (Aa)
    and nothing to ignore"""

    # arrange
    tmp_path = Path(os.path.dirname(share_it_monography))
    file_name= os.path.basename(share_it_monography)
    sch_mock.return_value = _place_file(tmp_path, file_name)

    # act
    outcome = ddb_validation_sch(share_it_monography)

    # assert
    assert sch_mock.call_count == 1
    assert len(outcome) == 0    # nothing invalid to see


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
@mock.patch('digiflow.digiflow_validate._forward_sch_cli')
def test_ddb_schematron_opendata_44046_no_ignores(sch_mock, share_it_monography):
    """Schematron validation with common monography (Aa)
    and nothing to ignore"""

    # arrange
    tmp_path = Path(os.path.dirname(share_it_monography))
    file_name= os.path.basename(share_it_monography)
    sch_mock.return_value = _place_file(tmp_path, file_name)

    # act
    outcome = ddb_validation_sch(share_it_monography, ignore_rules=[])

    # assert
    assert sch_mock.call_count == 1
    assert outcome['info'] == [('identifier_01', 'type=gbv')]


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_xslt_opendata_44046(share_it_monography):
    """XSLT validation with common monography (Aa)"""

    # act
    outcome = ddb_validation(share_it_monography, ignore_rules=[])

    # assert
    assert len(outcome) == 1
    assert 'identifier_01' in outcome[DDB_OTHER][0]
    assert '265982944' in outcome[DDB_OTHER][0]


@mock.patch('digiflow.digiflow_validate._forward_sch_cli')
@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
def test_ddb_schematron_opendata_44046_identifier(sch_mock, share_it_monography):
    """Suppress identifier validation message for 
    simple monography (Aa)"""

    # arrange
    tmp_path = Path(os.path.dirname(share_it_monography))
    file_name= os.path.basename(share_it_monography)
    sch_mock.return_value = _place_file(tmp_path, file_name)

    # act
    outcome = ddb_validation_sch(share_it_monography, ignore_rules=['identifier_01'])

    # assert
    assert sch_mock.call_count == 1
    assert 'info' not in outcome
    assert len(outcome) == 0


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
@mock.patch('digiflow.digiflow_validate._forward_sch_cli')
def test_ddb_schematron_opendata_44046_ignorances_as_string(sch_mock, share_it_monography):
    """Ensure also ignore information can be passed as string"""

    # arrange
    tmp_path = Path(os.path.dirname(share_it_monography))
    file_name= os.path.basename(share_it_monography)
    sch_mock.return_value = _place_file(tmp_path, file_name)

    # act
    outcome = ddb_validation_sch(share_it_monography, ignore_rules='identifier_01')

    # assert
    assert sch_mock.call_count == 1
    assert 'info' not in outcome
    assert len(outcome) == 0


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_xslt_opendata_44046_plain(share_it_monography):
    """Suppress identifier validation message for 
    simple monography (Aa)"""

    # act
    outcome = ddb_validation(share_it_monography, ignore_rules=['identifier_01'])

    # assert
    assert 'info' not in outcome
    assert len(outcome) == 0


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_xslt_validation_kitodo2_legacy_monography(tmp_path):
    """XSLT validation outcome for out-dated
    print export data from Kitodo2 legacy system"""

    # arrange
    file_name = 'k2_mets_vd18_147638674.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target)

    # assert
    _exc_load = _dexc.value.args[0]
    assert len(_exc_load) == 2
    assert isinstance(_exc_load, list)
    # changed from 05/22 => 06/23
    # from location_02 => location_01
    assert 'location_01' in _exc_load[0]
    assert 'dmdSec_04' in _exc_load[1]


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
def test_ddb_schematron_validation_kitodo2_legacy_monography(tmp_path):
    """Schematron validation outcome for out-dated
    print export data from Kitodo2 legacy system"""

    # arrange
    file_name = 'k2_mets_vd18_147638674.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    result = ddb_validation_sch(mets_target)

    # assert 
    assert len(result) == 2
    assert len(result['error']) == 2
    assert result['error'][0][0] == 'location_02'
    assert result['error'][1][0] == 'dmdSec_04'
    assert len(result['warn']) == 1
    assert result['warn'][0][0] == 'amdSec_13'  # should be PD-licence


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
def test_ddb_schematron_validation_kitodo2_legacy_monography_curated(tmp_path):
    """DDB compliant data with curated METS which doesn't 
    contain former additional dmdSec anymore so only
    a single warning regarding licences remains"""

    # arrange
    file_name = 'k2_mets_vd18_147638674_curated.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    result = ddb_validation_sch(mets_target)

    # assert 
    assert len(result) == 1
    assert 'error' not in result
    assert len(result['warn']) == 1 
    assert result['warn'][0][0] == 'amdSec_13'


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_xslt_validation_kitodo2_menadoc_44080924x(tmp_path):
    """XSLT validation outcome for rather
    recent digitized object from menadoc retro-digi
    
    changes between 05/22 and 06/23
    * increase number of errors from 3 to 16(!)
    * assert 'originInfo_06' in result[DDB_ERROR][0]
    * assert 'location_02' in result[DDB_ERROR][1]
    * assert 'dmdSec_04' in result[DDB_ERROR][2]
    """

    # arrange
    file_name = 'k2_mets_mena_44080924X.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target)

    # assert
    _load = _dexc.value.args[0]
    assert len(_load) == 16
    assert 'originInfo_06' in _load[0]
    assert 'location_01' in _load[1] # before: location_02
    assert 'dmdSec_04' in _load[2]


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
def test_ddb_schematron_validation_kitodo2_menadoc_44080924x(tmp_path):
    """Schematron validation outcome for rather
    recent digitized object from menadoc retro-digi"""

    # arrange
    file_name = 'k2_mets_mena_44080924X.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    result = ddb_validation_sch(mets_target)

    # assert 
    assert len(result) == 2
    assert len(result['error']) == 3    # 3 errors
    assert result['error'][0][0] == 'originInfo_06'
    assert result['error'][1][0] == 'location_02'
    assert result['error'][2][0] == 'dmdSec_04'
    assert len(result['warn']) == 3     # 3 warnings
    assert result['warn'][0][0] == 'titleInfo_08'


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_xslt_validation_kitodo2_vd18_153142537_raw(tmp_path):
    """XSLT validation outcome for VD18 c-stage
    without default customm ignore rules
    
    changed from 05/22 => 06/23
    * increased errors from 7 to 11
    """

    # arrange
    file_name = 'k2_mets_vd18_153142537.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target, ignore_rules=[])

    # assert 
    _load = _dexc.value.args[0]
    assert len(_load) == 11
    assert 'titleInfo_02' in _load[0]
    assert 'originInfo_06' in _load[1]
    assert 'originInfo_06' in _load[2]
    assert 'location_01' in _load[3]
    assert 'dmdSec_04' in _load[4]
    assert 'fileSec_02' in _load[5]
    assert 'structMapLogical_17' in _load[6]


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
@mock.patch('digiflow.digiflow_validate._forward_sch_cli')
def test_ddb_schematron_validation_kitodo2_vd18_153142537_raw(sch_mock, tmp_path):
    """Schematron validation outcome for VD18 F-Stage
    without respect to custom ignore rules"""

    # arrange
    file_name = 'k2_mets_vd18_153142537.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))
    sch_mock.return_value = _place_file(tmp_path, file_name)

    # act
    result = ddb_validation_sch(mets_target, ignore_rules=[])

    # assert 
    assert sch_mock.call_count == 1
    assert len(result) == 2
    assert len(result['error']) == 7    # 6 errors
    assert result['error'][0][0] == 'structMapLogical_17'
    assert result['error'][1][0] == 'fileSec_02'
    assert result['error'][2][0] == 'titleInfo_02'
    assert result['error'][3][0] == 'originInfo_06'
    assert result['error'][4][0] == 'originInfo_06'
    assert result['error'][5][0] == 'location_02'
    assert result['error'][6][0] == 'dmdSec_04'
    assert 'warn' not in result         # no warning
    assert len(result['info']) == 1     # 1 info
    assert result['info'] == [('identifier_01', 'type=eki')]


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
@mock.patch('digiflow.digiflow_validate._forward_sch_cli')
def test_ddb_schematron_validation_kitodo2_vd18_153142537_raw_fatals(sch_mock, tmp_path):
    """Schematron validation outcome for VD18 F-Stage
    without default ignore rules and without merging
    invalid category 'fatal' into 'error'"""

    # arrange
    file_name = 'k2_mets_vd18_153142537.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))
    sch_mock.return_value = _place_file(tmp_path, file_name)

    # act
    result = ddb_validation_sch(mets_target, aggregate_errors=False, ignore_rules=[])

    # assert 
    assert sch_mock.call_count == 1
    assert len(result) == 3
    assert len(result['error']) == 5    # 4 errors remain
    assert result['error'][0][0] == 'titleInfo_02'
    assert result['error'][1][0] == 'originInfo_06'
    assert result['error'][2][0] == 'originInfo_06'
    assert result['error'][3][0] == 'location_02'
    assert result['error'][4][0] == 'dmdSec_04'
    assert len(result['fatal']) == 2    # 2 fatals
    assert result['fatal'][0][0] == 'structMapLogical_17'
    assert result['fatal'][1][0] == 'fileSec_02'
    assert 'warn' not in result         # no warning
    assert len(result['info']) == 1     # 1 info
    assert result['info'][0][0] == 'identifier_01'


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
@mock.patch('digiflow.digiflow_validate._forward_sch_cli')
def test_ddb_schematron_validation_kitodo2_vd18_153142537(sch_mock, tmp_path):
    """Schematron validation outcome for VD18 F-Stage
    which considers our default ignore rules"""

    # arrange
    file_name = 'k2_mets_vd18_153142537.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))
    sch_mock.return_value = _place_file(tmp_path, file_name)

    # act
    result = ddb_validation_sch(mets_target)

    # assert 
    assert sch_mock.call_count == 1
    assert len(result) == 1
    assert len(result['error']) == 4    # 4 errors remain
    assert result['error'][0][0] == 'originInfo_06'
    assert result['error'][1][0] == 'originInfo_06'
    assert result['error'][2][0] == 'location_02'
    assert result['error'][3][0] == 'dmdSec_04'
    assert 'warn' not in result         # no warning
    assert 'info' not in result         # no info


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
def test_ddb_schematron_validation_kitodo2_vd18_153142537_curated(tmp_path):
    """DDB compliant with curated METS which doesn't contain
    the former additional dmdSec anymore and also ignores
    originInfo_06 issues"""

    # arrange
    file_name = 'k2_mets_vd18_153142537_curated.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))
    _ignore_them = DDB_IGNORE_RULES_MVW + ['originInfo_06']

    # act
    result = ddb_validation_sch(mets_target, ignore_rules=_ignore_them)

    # assert 
    assert len(result) == 0
    assert 'error' not in result    # no errors
    assert 'warn' not in result     # no warning
    assert 'info' not in result     # no info


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
@mock.patch('digiflow.digiflow_validate._forward_sch_cli')
def test_ddb_schematron_validation_kitodo2_vd18_153142340_raw(sch_mock, tmp_path):
    """Schematron validation outcome for VD18 C-Stage
    corresponding to k2_mets_153142537.xml as-it-is"""

    # arrange
    file_name = 'k2_mets_vd18_153142340.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))
    sch_mock.return_value = _place_file(tmp_path, file_name)

    # act
    result = ddb_validation_sch(mets_target, ignore_rules=[])

    # assert 
    assert sch_mock.call_count == 1
    assert len(result) == 3
    assert len(result['error']) == 3    # 3 errors
    assert result['error'][0][0] == 'structMapLogical_17'
    assert result['error'][1][0] == 'originInfo_06'
    assert result['error'][2][0] == 'originInfo_06'
    assert len(result['warn']) == 1     # 1 warning
    assert result['warn'][0][0] == 'location_03'    # missing physicalLocation => but it's c-stage?
    assert len(result['info']) == 1     # 1 info
    assert result['info'] == [('identifier_01', 'type=eki')]


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
@mock.patch('digiflow.digiflow_validate._forward_sch_cli')
def test_ddb_schematron_validation_kitodo2_vd18_153142340(sch_mock, tmp_path):
    """Schematron validation outcome for VD18 C-Stage
    corresponding to k2_mets_153142537.xml
    considering ULB default ignore rules"""

    # arrange
    file_name = 'k2_mets_vd18_153142340.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))
    sch_mock.return_value = _place_file(tmp_path, file_name)

    # act
    result = ddb_validation_sch(mets_target)

    # assert 
    assert sch_mock.call_count == 1
    assert len(result) == 2
    assert len(result['error']) == 2    # 2 errors remain
    assert result['error'][0][0] == 'originInfo_06'
    assert result['error'][1][0] == 'originInfo_06'
    assert len(result['warn']) == 1     # 1 warning
    assert result['warn'][0][0] == 'location_03'
    assert 'info' not in result     # no info


@pytest.mark.skipif(not os.path.isfile(SCHEMATRON_BIN), 
    reason=f'no schematron binary at {SCHEMATRON_BIN}')
def test_ddb_schematron_validation_kitodo2_morbio_1748529021(tmp_path):
    """Schematron validation outcome for recent Morbio
    export considering ULB default ignore rules"""

    # arrange
    file_name = 'k2_mets_morbio_1748529021.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    result = ddb_validation_sch(mets_target)

    # assert 
    assert len(result) == 2
    assert len(result['error']) == 3    # 3 errors
    assert result['error'][0][0] == 'location_02'
    assert result['error'][1][0] == 'dmdSec_04' # the suspicious additional dmdSec
    assert result['error'][2][0] == 'amdSec_05' # contains cc-by-3 => not accepted
    assert 'warn' not in result     # no warnings
    assert len(result['info']) == 2 # 2 info
    assert result['info'][0][0] == 'subject_01' # subject/topic without gnd-link "Sonstige..."
    assert result['info'][1][0] == 'subject_01' # subject/topic without gnd-link "Notarsurkunde"


def test_ddb_validate_newspaper(tmp_path):
    """Test recent DDB-Validation 06/23 for
    migrated newspaper issue missing
    * license information
    """

    # arrange
    the_name = 'vls_digital_3014754.zmets.xml'
    mets_source = Path(TEST_RES) / the_name
    mets_target = Path(str(tmp_path), the_name)
    shutil.copy(str(mets_source), str(mets_target))

     # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target, digi_type='issue')

    # assert
    _load = _dexc.value.args[0]
    assert len(_load) == 1
    assert '[amdSec_04]' in _load[0]



# -*- coding: utf-8 -*-

import os
import shutil

import lxml.etree as ET

import pytest

from digiflow import (
    generate_urn_pz,
    enrich_urn_kitodo2,
    enrich_urn_granular,
    write_xml_file,
    does_fit,
    URNType,
    GranularURNExistsException,
    MetsReader,
    XMLNS
)

from .conftest import (
    TEST_RES
)

LABEL_META = 'meta.xml'
XPATH_GOOBI_URN = './/goobi:metadata[@name="ulb_urn"]'


@pytest.mark.parametrize(
    "test_input,expected", [
        ('urn:nbn:de:gbv:089-332175294', '5'),
        ('urn:nbn:de:gbv:3:1-1192015415-181497433-1', '9'),
        ('urn:nbn:de:gbv:3:4-1192015415-211620807-1', '7'),
        ('urn:nbn:de:gbv:3:1-62923', '0'),
        ('urn:nbn:de:gbv:3:1-69482', '1'),
        ('urn:nbn:de:gbv:3:3-21437-p0004-', '6'),
        ('urn:nbn:de:gbv:3:1-847518-1901041801', '8')])
def test_urn_checkdigits(test_input, expected):
    """
    Checkdigit algorithm (Pruefziffer)

    http://www.persistent-identifier.de/?link=316
    http://nbn-resolving.de/nbnpruefziffer.php

    ULB Schema: urn:nbn:de:gbv:3:<Sammlung>-<SystemId>-<PPN>-<ExemplarNr>
        (see: https://redmine.itz.uni-halle.de/projects/ulb-dd/wiki/URN_Generierung_Kitodo)
    """

    assert expected == generate_urn_pz(test_input)


@pytest.mark.parametrize(
    "test_input,expected", [
        ('urn:nbn:de:gbv:089-332175294-5', False),
        ('urn:nbn:de:gbv:3:1-1192015415-181497433-19', True),
        ('urn:nbn:de:gbv:n.a.', False),
        ('urn:nbn:de:gbv:3:3-21437-p0004-6', True),
        ('urn:nbn:de:gbv:3:1-694468/fragment/page=9293732', True),
        ('urn:nbn:de:gbv:3:1-847518-19010418018', True),
        ('urn:nbn:de:gbv:3:1-847518-19010418018/fragment/page=0001', True)])
def test_existing_urn_values(test_input, expected, tmp_path):
    """
    Which known URN patters are recognized as
    matching URNs?

    Please note: Granular page-related URN 2.0 are
    considered as valid URNS in this case because they are
    composed of a valid URN (with check digit)
    and following fragment extension
    """

    # arrange
    src_path = TEST_RES / "k2_mets_vd18_147638674.xml"
    path_meta = tmp_path / "147638674.xml"
    shutil.copyfile(src_path, str(path_meta))
    xml_root = ET.parse(path_meta).getroot()
    parent_phys = xml_root.find('.//mets:structMap[@TYPE="PHYSICAL"]', XMLNS)
    dummy_element = ET.SubElement(parent_phys, 'div', {'CONTENTIDS': test_input})

    # assert
    assert expected == does_fit(dummy_element)


@pytest.fixture
def fixture_k2_export(tmp_path):
    src_path = TEST_RES / "k2_mets_vd18_147638674.xml"
    path_meta = tmp_path / "147638674.xml"
    shutil.copyfile(src_path, str(path_meta))
    return str(path_meta)


def test_urn_granular_exported_from_kitodo2_default(fixture_k2_export):
    """Ensure padd_left is respected"""

    # arrange
    _urn_main = 'urn:nbn:de:gbv:3:1-1192015415-147638674-17'
    xml_root = ET.parse(fixture_k2_export).getroot()

    # act
    (xml_root_result, kind, report) = enrich_urn_granular(xml_root,
                                                       urn_main=_urn_main,
                                                       padd_left=1)
    write_xml_file(xml_root_result, fixture_k2_export)

    # assert
    assert len(report.added) == 17
    assert kind == URNType.GRANULAR_FRAGMENT_PAGE
    result_xml = ET.parse(fixture_k2_export).getroot()
    phys_divs = result_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]/mets:div/mets:div', XMLNS)
    for phys_div in phys_divs:
        assert phys_div.attrib['CONTENTIDS']

    urn = f'{_urn_main}/fragment/page=1'
    assert urn == phys_divs[0].attrib['CONTENTIDS']
    urn2 = f'{_urn_main}/fragment/page=17'
    assert urn2 == phys_divs[16].attrib['CONTENTIDS']


def test_urn_granular_exported_with_padding(fixture_k2_export):
    """Behavior if page padding is fixed to '4'"""

    # arrange
    _urn_main = 'urn:nbn:de:gbv:3:1-1192015415-147638674-17'
    if not os.path.exists(fixture_k2_export):
        raise ValueError(f'invalid path "{fixture_k2_export}"')
    xml_root = ET.parse(fixture_k2_export).getroot()

    # act
    (xml_root_result, kind, rep) = enrich_urn_granular(xml_root, _urn_main, padd_left=4)
    write_xml_file(xml_root_result, fixture_k2_export)

    # assert
    assert len(rep.added) == 17
    assert kind == URNType.GRANULAR_FRAGMENT_PAGE
    result_xml = ET.parse(fixture_k2_export).getroot()
    phys_divs = result_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]/mets:div/mets:div', XMLNS)
    urn = f'{_urn_main}/fragment/page=0001'
    assert urn == phys_divs[0].attrib['CONTENTIDS']
    urn2 = f'{_urn_main}/fragment/page=0017'
    assert urn2 == phys_divs[16].attrib['CONTENTIDS']


@pytest.fixture
def fixture_k2_plain_monography(tmp_path):
    src_path = TEST_RES / "kitodo2-goobi" / "monography" / "meta_empty_structs.xml"
    path_meta = tmp_path / LABEL_META
    shutil.copyfile(src_path, str(path_meta))
    return str(path_meta)


def test_kitodo2_monography_enrich_urn(fixture_k2_plain_monography):
    """Ensure expected URN generated and persisted"""

    # act
    result = enrich_urn_kitodo2(fixture_k2_plain_monography)

    # assert
    assert len(result) == 1
    result_xml = ET.parse(fixture_k2_plain_monography).getroot()
    urn = result_xml.find(XPATH_GOOBI_URN, XMLNS)
    assert "urn:nbn:de:gbv:3:1-1192015415-269210156-14" == urn.text


@pytest.fixture
def fixture_k2_monography(tmp_path):
    src_path = TEST_RES / "kitodo2-goobi" / "monography" / "meta.xml"
    path_meta = tmp_path / LABEL_META
    shutil.copyfile(src_path, str(path_meta))
    return str(path_meta)


def test_create_urn_granular_within_kitodo2(fixture_k2_monography):

    # arrange
    _urn_main = 'urn:nbn:de:gbv:3:3-1192015415-882674978-12'
    if not os.path.exists(fixture_k2_monography):
        raise ValueError(f'invalid path "{fixture_k2_monography}"')
    xml_root = ET.parse(fixture_k2_monography).getroot()

    # act
    (xml_root_result, kind, rep) = enrich_urn_granular(xml_root, urn_main =_urn_main, padd_left=8)
    write_xml_file(xml_root_result, fixture_k2_monography)

    # assert
    assert len(rep.added) == 27
    assert kind == URNType.GRANULAR_FRAGMENT_PAGE
    result_xml = ET.parse(fixture_k2_monography).getroot()
    phys_divs = result_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]/mets:div/mets:div', XMLNS)
    for phys_div in phys_divs:
        assert phys_div.attrib['CONTENTIDS']

    urn = f'{_urn_main}/fragment/page=00000001'
    assert urn == phys_divs[0].attrib['CONTENTIDS']
    urn1 = f'{_urn_main}/fragment/page=00000017'
    assert urn1 == phys_divs[16].attrib['CONTENTIDS']


def test_urn_mvw_kitodo2(tmp_path):
    """Insert expected URNs both for volume and host"""

    # arrange
    src_meta = TEST_RES / "kitodo2-goobi" / "mvw" / "meta.xml"
    path_meta = tmp_path / LABEL_META
    shutil.copyfile(src_meta, str(path_meta))
    src_anchor = TEST_RES / "kitodo2-goobi" / "mvw" / "meta_anchor.xml"
    path_anchor = tmp_path / "meta_anchor.xml"
    shutil.copyfile(src_anchor, str(path_anchor))
    dir_name = os.path.dirname(path_meta)

    # act
    enrich_urn_kitodo2(path_meta)

    # assert
    meta_xml = ET.parse(path_meta).getroot()
    urn = meta_xml.find(XPATH_GOOBI_URN, XMLNS)
    assert "urn:nbn:de:gbv:3:1-1192015415-325168768-18" == urn.text
    anchor_xml = ET.parse(os.path.join(
        dir_name, 'meta_anchor.xml')).getroot()
    urn_host = anchor_xml.find(XPATH_GOOBI_URN, XMLNS)
    assert "urn:nbn:de:gbv:3:1-1192015415-147173272-15" == urn_host.text


@pytest.fixture
def fixture_vls_monography_granular1(tmp_path):
    src_meta = TEST_RES / "vls" / "monography" / "737429-migrated.xml"
    path_meta = tmp_path / "737429.xml"
    shutil.copyfile(src_meta, str(path_meta))
    return str(path_meta)


def test_urn_granular_upsert_legacy_vls(fixture_vls_monography_granular1):
    """
    Ensure: Digitalizates with existings granular URNs preserve
    their URN-Type and that granular URNs are added if missing on
    physical containers
    """

    # arrange
    urn_main = 'urn:nbn:de:gbv:3:3-21437'
    xml_root = ET.parse(fixture_vls_monography_granular1).getroot()

    # act
    (xml_root_result, _type, rep) = enrich_urn_granular(xml_root, urn_main=urn_main)
    write_xml_file(xml_root_result, fixture_vls_monography_granular1)

    # assert
    assert os.path.exists(fixture_vls_monography_granular1)
    assert len(rep.added) == 1
    assert _type == URNType.GRANULAR_PAGE
    meta_xml = ET.parse(fixture_vls_monography_granular1).getroot()
    page5 = meta_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]//mets:div[@CONTENTIDS]', XMLNS)[4]
    assert "urn:nbn:de:gbv:3:3-21437-p0005-2" == page5.attrib['CONTENTIDS']


def test_urn_granular_with_inconsistent_contentids(tmp_path):
    """
    ensure that inconsistent legacy data can be partially be
    helped by adding a new granular urn which would collide
    with an existing urn

    http://digital.bibliothek.uni-halle.de/hd/oai/?verb=GetRecord&metadataPrefix=mets&mode=xml&identifier=10595
    """

    # arrange
    src_meta = TEST_RES / "vls" / "monography" / "10595.xml"
    path_meta = tmp_path / "10595.xml"
    shutil.copyfile(src_meta, str(path_meta))
    xml_root = ET.parse(str(path_meta)).getroot()

    # act
    with pytest.raises(GranularURNExistsException) as execinfo:
        enrich_urn_granular(xml_root, urn_main='urn:nbn:de:gbv:3:3-178')
        assert 'calculated granular URN ' in execinfo.value


def test_urn_granular_vd18_mvw_fstage(tmp_path):
    """
    ensure that existing granular 1.0 is preserved
    """

    # arrange
    main_urn = 'urn:nbn:de:gbv:3:1-635986'
    src_meta = TEST_RES / "migration" / "9427337.mets.xml"
    path_meta = tmp_path / "9427337.mets.xml"
    shutil.copyfile(src_meta, str(path_meta))
    xml_root = ET.parse(str(path_meta)).getroot()

    # act
    (xml_root_result, _type, rep) = enrich_urn_granular(xml_root, main_urn)
    write_xml_file(xml_root_result, str(path_meta))

    # assert
    assert os.path.exists(str(path_meta))
    assert len(rep.added) == 0
    assert _type == URNType.GRANULAR_PAGE
    meta_xml = ET.parse(str(path_meta)).getroot()
    page5 = meta_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]//mets:div[@CONTENTIDS]', XMLNS)[1]
    assert "urn:nbn:de:gbv:3:1-635986-p0002-0" == page5.attrib['CONTENTIDS']


@pytest.fixture
def fixture_granular_urn_order_gap(tmp_path):
    """
    create complex invalid example with legacy data, where a granular URN
    seems to be missing in between pages, like it jumps from p0001 to p0003:
    """

    src_meta = TEST_RES / "vls" /"12504.mets.xml"  # mets with order gaps
    path_meta = tmp_path / "12504.mets.xml"
    shutil.copyfile(src_meta, str(path_meta))
    xml_root = ET.parse(str(path_meta)).getroot()
    last_phys_cnt = xml_root.find('.//mets:*[@ID="phys58793"]', XMLNS)
    parent = last_phys_cnt.getparent()
    # 2. insert brand new colorchecker as last element
    attribs = {
        'ID': 'phys58794',
        'TYPE': 'page',
        'CONTENTIDS': 'urn:n.a.',
        'ORDER': '13',
        'ORDERLABEL': '[Colorchecker]'}
    ET.SubElement(parent, '{http://www.loc.gov/METS/}div', attribs)
    write_xml_file(xml_root, path_meta)
    return str(path_meta), xml_root


def test_urn_granular_order_gaps_repair(
        tmp_path, fixture_granular_urn_order_gap):
    """
    Behavior for inconsistent legacy data with gaps
    in their ORDER

    11 pages do already have a granular URN, the last one misses it
    Since the ORDER-attribute leaps from "1" to "3" for second page,
    it can't use the regular way to count the last ORDER-attribute from
    last page up by one, since the resulting URN is already owned
    by the pre-last page
    """
    # arrange
    main_urn = 'urn:nbn:de:gbv:3:1-1488'
    xml_root = fixture_granular_urn_order_gap[1]
    valid_urns_at_start = 12

    # pre-check
    cnt_ids = xml_root.findall('.//mets:structMap[@TYPE="PHYSICAL"]//mets:div[@ORDER]', XMLNS)
    assert len([e for e in cnt_ids if '-p00' in e.get('CONTENTIDS')]) == valid_urns_at_start

    # act I
    with pytest.raises(GranularURNExistsException) as exist_exc:
        enrich_urn_granular(xml_root, main_urn)

    # assert
    assert "granular URN 'urn:nbn:de:gbv:3:1-1488-p0013-5' already used" in str(exist_exc.value)

    # act II: try to repair this time
    # (xml_res, _, ins) = enrich_urn_granular_from_contentids(fixture_granular_urn_order_gap[1])
    (xml_res, _, rep) = enrich_urn_granular(fixture_granular_urn_order_gap[1], urn_main=main_urn, do_sanitize=True)
    path_meta = tmp_path / "12504_repair.mets.xml"
    write_xml_file(xml_res, str(path_meta))
    colorchecker = xml_res.find('.//mets:*[@ID="phys58794"]', XMLNS)
    attr = colorchecker.attrib

    # assert
    assert len(rep.added) == 1
    assert attr['CONTENTIDS'] == 'urn:nbn:de:gbv:3:1-1488-p0014-1'
    assert attr['ORDER'] == '13'
    assert attr['ORDERLABEL'] == '[Colorchecker]'

    # act
    cnt2 = xml_root.findall('.//mets:structMap[@TYPE="PHYSICAL"]//mets:div[@ORDER]', XMLNS)

    # assert
    assert len([e for e in cnt2 if '-1488-p00' in e.attrib['CONTENTIDS']]) == (valid_urns_at_start + 1)


@pytest.fixture
def fixture_granular_urn_false_order(tmp_path):
    """
    create complex invalid example with legacy data, where a granular URN
    seems to be missing in between pages, like it jumps from p0001 to p0003:
    """
    # mets with order mismatch
    src_meta = TEST_RES / "vls" / "4066583.mets.xml"
    path_meta = tmp_path / "4066583.mets.xml"
    shutil.copyfile(src_meta, str(path_meta))
    xml_root = ET.parse(str(path_meta)).getroot()
    last_phys_cnt = xml_root.find('.//mets:*[@ID="phys4116796"]', XMLNS)
    parent = last_phys_cnt.getparent()
    # 2. insert brand new colorchecker as last element
    attribs = {
        'ID': 'phys4116797',
        'TYPE': 'page',
        'CONTENTIDS': 'urn:n.a.',
        'ORDER': '71',
        'ORDERLABEL': '[Colorchecker]'}
    ET.SubElement(parent, '{http://www.loc.gov/METS/}div', attribs)

    return str(path_meta), xml_root


def test_urn_granular_false_order(
        tmp_path, fixture_granular_urn_false_order):
    """
    check behavior for inconsistent legacy with gaps in in granular urns
    and partly missorted urn
    when new container is attached to end, like it is with colorchecker
    data
    """
    # arrange
    xml_root = fixture_granular_urn_false_order[1]
    n_valid_urns = 70

    # pre-check
    cnt_ids = xml_root.findall('.//mets:structMap[@TYPE="PHYSICAL"]//mets:div[@ORDER]', XMLNS)
    urns = [e for e in cnt_ids if '-272189-p00' in e.attrib['CONTENTIDS']]
    assert len(urns) == n_valid_urns

    # act II: try to repair this time
    xml_res, _, rep = enrich_urn_granular(
        fixture_granular_urn_false_order[1],
        'urn:nbn:de:gbv:3:1-272189', do_sanitize=True)
    path_meta = tmp_path / "4066583_repair.mets.xml"
    write_xml_file(xml_res, str(path_meta))
    colorchecker = xml_res.find('.//mets:*[@ID="phys4116797"]', XMLNS)
    attr = colorchecker.attrib

    # assert
    assert len(rep.added) == 1
    assert attr['CONTENTIDS'] == 'urn:nbn:de:gbv:3:1-272189-p0072-1'
    assert attr['ORDER'] == '71'
    assert attr['ORDERLABEL'] == '[Colorchecker]'

    # act
    cnt2 = xml_root.findall('.//mets:structMap[@TYPE="PHYSICAL"]//mets:div[@ORDER]', XMLNS)

    # assert after enrich
    urns = [e for e in cnt2 if '-272189-p00' in e.attrib['CONTENTIDS']]
    assert len(urns) == (n_valid_urns + 1)


def test_urn_granular_fragement_page_k3_newspaper_issue(tmp_path):
    """
    ensure that Kitodo3 Newspaper issue gets
    granular URNs in expected order, i.e.:
    first page = [Seite 1]
    last page  = [Colorchecker]
    """

    # arrange
    urn_main = 'urn:nbn:de:gbv:3:3-171133730-30089663818490701-14'
    src_meta = TEST_RES / "k3_300896638-18490701.xml"
    path_meta = tmp_path / "300896638-18490701.mets.xml"
    shutil.copyfile(src_meta, str(path_meta))
    xml_root = ET.parse(str(path_meta)).getroot()

    # act
    (xml_root_result, _type, rep) = enrich_urn_granular(xml_root, urn_main)
    write_xml_file(xml_root_result, str(path_meta))

    # assert
    assert os.path.exists(str(path_meta))
    assert len(rep.added) == 5
    assert _type == URNType.GRANULAR_FRAGMENT_PAGE
    meta_xml = ET.parse(str(path_meta)).getroot()
    fst_page = meta_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]//mets:div[@CONTENTIDS]', XMLNS)[0]
    assert fst_page.attrib['CONTENTIDS'].endswith("30089663818490701-14/fragment/page=0001")
    assert fst_page.attrib['ORDERLABEL'] == '[Seite 1]'
    lst_page = meta_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]//mets:div[@CONTENTIDS]', XMLNS)[-1]
    assert lst_page.attrib['CONTENTIDS'].endswith("30089663818490701-14/fragment/page=0005")
    assert lst_page.attrib['ORDERLABEL'] == '[Colorchecker]'


def test_urn_granular_inhouse_zd_replacement_not_enforced(tmp_path):
    """
    ensure existing granular urns are
    kept if no replacement enforced
    """

    # arrange
    urn_main = 'urn:nbn:de:gbv:3:3-62299-19030423018'
    src_meta = TEST_RES / "vls_digital_3014754.zmets.xml"
    path_meta = tmp_path / "3014754.mets.xml"
    shutil.copyfile(src_meta, str(path_meta))
    _reader = MetsReader(path_meta)
    ident_urn = ET.SubElement(_reader.primary_dmd, '{http://www.loc.gov/mods/v3}identifier', {'type': 'urn'})
    ident_urn.text = urn_main
    _reader.write()
    xml_root = ET.parse(str(path_meta)).getroot()

    # act
    (xml_root_result, _type, rep) = enrich_urn_granular(xml_root, urn_main)
    write_xml_file(xml_root_result, str(path_meta))

    # assert
    assert os.path.exists(str(path_meta))
    assert len(rep.added) == 0
    assert len(rep.existed) == 4
    assert _type == URNType.GRANULAR_PAGE
    meta_xml = ET.parse(str(path_meta)).getroot()
    page1 = meta_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]//mets:div[@CONTENTIDS]', XMLNS)[0]
    assert "urn:nbn:de:gbv:3:3-62299-p0005-5" == page1.attrib['CONTENTIDS']


def test_urn_granular_inhouse_zd_replacement_enforced(tmp_path):
    """
    ensure existing granular urns can be exchanged
    if re-creation enforced
    """

    # arrange
    urn_main = 'urn:nbn:de:gbv:3:3-62299-19030423018'
    src_meta = TEST_RES / "vls_digital_3014754.zmets.xml"
    path_meta = tmp_path / "3014754.mets.xml"
    shutil.copyfile(src_meta, str(path_meta))
    _reader = MetsReader(path_meta)
    ident_urn = ET.SubElement(_reader.primary_dmd, '{http://www.loc.gov/mods/v3}identifier', {'type': 'urn'})
    ident_urn.text = urn_main
    _reader.write()
    xml_root = ET.parse(str(path_meta)).getroot()

    # act
    (xml_root_result, _type, rep) = enrich_urn_granular(xml_root, urn_main=urn_main, do_replace=True)
    write_xml_file(xml_root_result, str(path_meta))

    # assert
    assert os.path.exists(str(path_meta))
    assert len(rep.added) == 4
    assert len(rep.replaced) == 4
    assert _type == URNType.GRANULAR_FRAGMENT_PAGE
    meta_xml = ET.parse(str(path_meta)).getroot()
    page1 = meta_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]//mets:div[@CONTENTIDS]', XMLNS)[0]
    assert "urn:nbn:de:gbv:3:3-62299-19030423018/fragment/page=0001" == page1.attrib['CONTENTIDS']


def test_urn_granular_zdp_replace_granular_urn(tmp_path):
    """
    Strange constellation: pages got granular URN related to
    newspaper's year, but the issues with the page got no
    particular granular URN themselves

    Issue: 1840-12-31

    PLEASE NOTE:
        Since the existing granular URN do *not* fit the
        expected URN pattern, there's actual no replacement
        taking place, therefore do_replace param can
        be ommitted
    """

    # arrange
    urn_main = 'urn:nbn:de:gbv:3:1-1823278450-18401231'
    src_meta = TEST_RES / "vls_digitale_9633116.zmets.xml"
    path_meta = tmp_path / "9633116.zmets.xml"
    shutil.copyfile(src_meta, str(path_meta))
    _reader = MetsReader(path_meta)
    ident_urn = ET.SubElement(_reader.primary_dmd, '{http://www.loc.gov/mods/v3}identifier', {'type': 'urn'})
    ident_urn.text = urn_main
    _reader.write()
    xml_root = ET.parse(str(path_meta)).getroot()

    # act
    (xml_root_result, _type, rep) = enrich_urn_granular(xml_root, urn_main=urn_main, do_replace=True)
    write_xml_file(xml_root_result, str(path_meta))

    # assert
    assert os.path.exists(str(path_meta))
    assert len(rep.added) == 6
    assert len(rep.existed) == 6
    assert rep.existed[0].urn == 'urn:nbn:de:gbv:3:1-694468/fragment/page=9293732'
    assert rep.added[0].urn == 'urn:nbn:de:gbv:3:1-1823278450-18401231/fragment/page=0001'
    assert len(rep.replaced) == 6
    assert rep.replaced['phys9293732'] == 'urn:nbn:de:gbv:3:1-694468/fragment/page=9293732 => urn:nbn:de:gbv:3:1-1823278450-18401231/fragment/page=0001'
    assert rep.existed[0].id == rep.added[0].id
    assert _type == URNType.GRANULAR_FRAGMENT_PAGE
    meta_xml = ET.parse(str(path_meta)).getroot()
    page1 = meta_xml.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]//mets:div[@CONTENTIDS]', XMLNS)[0]
    assert "urn:nbn:de:gbv:3:1-1823278450-18401231/fragment/page=0001" == page1.attrib['CONTENTIDS']


def test_get_identifiers_from_elder_datasets(tmp_path):
    """
    Ensure working impl even on elder data sets
    """

    # arrange
    src_meta = TEST_RES / "vls_digital_737429.mets.xml"
    path_meta = tmp_path / "737429.mets.xml"
    shutil.copyfile(src_meta, str(path_meta))
    _reader = MetsReader(path_meta, dmd_id='md737429')

    # act
    report = _reader.report

    # assert
    assert report.system_identifiers == {'digital.bibliothek.uni-halle.de/hd': '737429'}



"""Metadata module regarding METS/MODS"""

# -*- coding: utf-8 -*-
import abc
import os
import time

from collections import (
    defaultdict
)
from configparser import (
    ConfigParser
)
from pathlib import (
    Path
)
from typing import (
    List,
    Optional,
    Tuple
)

from lxml import etree as ET


# old-school: register all namespaces known so-far for writing
ET.register_namespace('dc', 'http://purl.org/dc/elements/1.1/')
ET.register_namespace('dv', 'http://dfg-viewer.de/')
ET.register_namespace('epicur', 'urn:nbn:de:1111-2004033116')
ET.register_namespace('marcxml', 'http://www.loc.gov/MARC21/slim')
ET.register_namespace('mets', 'http://www.loc.gov/METS/')
ET.register_namespace('mods', 'http://www.loc.gov/mods/v3')
ET.register_namespace('oai', 'http://www.openarchives.org/OAI/2.0/')
ET.register_namespace('oai_dc', 'http://www.openarchives.org/OAI/2.0/oai_dc/')
ET.register_namespace('ulb', 'https://bibliothek.uni-halle.de')
ET.register_namespace('vl', 'http://visuallibrary.net/vl')
ET.register_namespace('vls', 'http://semantics.de/vls')
ET.register_namespace('vlz', 'http://visuallibrary.net/vlz/1.0/')
ET.register_namespace('xlink', 'http://www.w3.org/1999/xlink')
ET.register_namespace('zvdd', 'http://zvdd.gdz-cms.de/')


XMLNS = {
    'alto'   : 'http://www.loc.gov/standards/alto/ns-v4#',
    'dc'     : 'http://purl.org/dc/elements/1.1/',
    'dv'     : 'http://dfg-viewer.de/',
    'epicur' : 'urn:nbn:de:1111-2004033116',
    'marcxml': 'http://www.loc.gov/MARC21/slim',
    'goobi'  : 'http://meta.goobi.org/v1.5.1/',
    'mets'   : 'http://www.loc.gov/METS/',
    'mix'    : 'http://www.loc.gov/mix/v20',
    'mods'   : 'http://www.loc.gov/mods/v3',
    'oai'    : 'http://www.openarchives.org/OAI/2.0/',
    'oai_dc' : 'http://www.openarchives.org/OAI/2.0/oai_dc/',
    'ulb'    : 'https://bibliothek.uni-halle.de',
    'vl'     : 'http://visuallibrary.net/vl',
    'vlz'    : 'http://visuallibrary.net/vlz/1.0/',
    'xlink'  : 'http://www.w3.org/1999/xlink',
    'zvdd'   : 'http://zvdd.gdz-cms.de/',
}

#
# METS/MODS common python-flavour XML expressions / namespaces attributes
#
XLINK_HREF = '{http://www.w3.org/1999/xlink}href'
XLINK_FROM = '{http://www.w3.org/1999/xlink}from'
XLINK_TO = '{http://www.w3.org/1999/xlink}to'
XPATH_STRUCTMAP_LOG = './/mets:structMap[@TYPE="LOGICAL"]/mets:div'
XPATH_MODS_VL_PICA = './/mods:extension/vlz:externalType[@recordSyntax="pica"]'
MARK_KITODO2 = 'kitodo2'
MARK_KITODO3 = 'kitodo3'
MARK_VLS = 'vls'
MARK_AGENT_LEGACY = 'legacy vlid: '
MARK_AGENT_VLID = 'created vlid: '
XPR_MODS_SEC = './/mets:dmdSec/mets:mdWrap/mets:xmlData/mods:mods/../../..'
PATTERN_FILEGROUP_USE = './/mets:fileGrp[@USE="{}"]'

# script constants
DEFAULT_FLAT_STRUCTS = [
    'monograph', 'manuscript'
]
DEFAULT_PARENT_STRUCTS = [
    'multivolume', 'multivolume_work', 'multivolumework',
    'fulltome', 'newspaper', 'year', 'periodical'
]
CONTENT_FILE_GROUPS = [
    'MAX', 'LOCAL'
]

# well grounded script root
DIGIFLOW_METADATA_ROOT = os.path.dirname(os.path.abspath(__file__))

# schema validtion XSDs
METS_1_12 = os.path.join(DIGIFLOW_METADATA_ROOT,
                         'resources', 'xsd', 'mets_1-12.xsd')
MODS_3_7 = os.path.join(DIGIFLOW_METADATA_ROOT,
                        'resources', 'xsd', 'mods_3-7.xsd')
MIX_2_0 = os.path.join(DIGIFLOW_METADATA_ROOT,
                       'resources', 'xsd', 'mix_2-0.xsd')
ALTO_3_1 = os.path.join(DIGIFLOW_METADATA_ROOT,
                        'resources', 'xsd', 'alto_3-1.xsd')
ALTO_4_2 = os.path.join(DIGIFLOW_METADATA_ROOT,
                        'resources', 'xsd', 'alto_4-2.xsd')
XSD_MAPPINGS = {'mets:mets': [METS_1_12], 'mods:mods': [MODS_3_7], 'mix:mix': [
    MIX_2_0], 'alto': [ALTO_4_2]}



def __is_schema_root(xml_tree, schema) -> bool:
    """
    Schema *might* be prefixed, like mets:mets, *or* not like ALTO-files
    Therefore we go for QName
    """
    qualified_name = ET.QName(xml_tree)
    local_name = qualified_name.localname
    return local_name in schema


def _is_contained(xml_tree, schema):
    if __is_schema_root(xml_tree, schema):
        return True
    return len(xml_tree.findall('.//' + schema, XMLNS)) > 0


def __validation_error(xmlschema, xml_tree):
    err = xmlschema.error_log
    raise RuntimeError(f"invalid schema '{xml_tree}' {err}")


def __validate_subtrees(xml_tree, schema, schema_tree):
    sections = xml_tree.findall('.//' + schema, XMLNS)
    for section in sections:
        if not schema_tree.validate(section):
            return __validation_error(schema_tree, xml_tree)


def _validate_with_xsd(xml_tree, schema, xsd_file):
    the_tree = ET.parse(xsd_file)
    schema_tree = ET.XMLSchema(the_tree)
    # Plese note:
    # going this way makes a vast difference from schema_tree.assertValidate !
    if __is_schema_root(
            xml_tree, schema) and not schema_tree.validate(xml_tree):
        return __validation_error(schema_tree, xml_tree)
    else:
        __validate_subtrees(xml_tree, schema, schema_tree)


def validate_xml(xml_data, xsd_mappings=XSD_MAPPINGS) -> bool:
    """
    Validate XML data with a set of given schema definitions (XSDs)

    :param xml_data: can be a str representing a local path, a PosixPath or ET.etree.root

    """

    # foster
    if isinstance(xml_data, Path):
        xml_data = str(xml_data)
    if isinstance(xml_data, str):
        xml_data = ET.parse(xml_data).getroot()

    # inspect
    for schema, xsd_files in xsd_mappings.items():
        if _is_contained(xml_data, schema):
            for xsd_file in xsd_files:
                _validate_with_xsd(xml_data, schema, xsd_file)

    # return default True if no error so far
    return True


def write_xml_file(xml_root,
                   outfile,
                   preamble='<?xml version="1.0" encoding="UTF-8"?>'):
    """write xml root prettified to outfile
    with default preamble for file usage
    create export dir if not exists and writeable
    """

    _prettified = pretty_xml(xml_root, preamble)
    dst_dir = os.path.dirname(outfile)
    if not os.path.exists(dst_dir):
        os.makedirs(dst_dir)
    with open(outfile, 'wb') as file_handler:
        file_handler.write(_prettified)


def pretty_xml(xml_root, preamble='<?xml version="1.0" encoding="UTF-8"?>'):
    """XML root as prettified string
    witd cleared namespace declarations
    and default preamble like for XML files

    disable preamble by setting it to 'None'
    """
    _as_string = ET.tostring(ET.ElementTree(xml_root), pretty_print=True, encoding='UTF-8')
    _pretty_parser = ET.XMLParser(resolve_entities=False,
                                 strip_cdata=False,
                                 remove_blank_text=True)
    _root = ET.fromstring(_as_string, _pretty_parser)
    ET.cleanup_namespaces(_root, top_nsmap=XMLNS)
    _formatted = ET.tostring(_root, pretty_print=True, encoding='UTF-8').decode('UTF-8')
    if preamble:
        formatted_file_content = f'{preamble}\n{_formatted}'
    else:
        formatted_file_content = f'{_formatted}'
    return formatted_file_content.encode('UTF-8').replace(b'\n', b'\r\n')


class XMLProcessor(abc.ABC):
    """Basic XML-Processing"""

    def __init__(self, path_xml, xmlns=None):
        self.tree = None
        if not os.path.exists(path_xml):
            raise RuntimeError(f"{path_xml} not existing!")
        self.path_xml = path_xml
        self.xmlns = xmlns
        if not self.xmlns:
            self.xmlns = XMLNS
        self.path_xml_dir = os.path.dirname(path_xml)
        self._parse()

    def _parse(self):
        self.tree = ET.parse(self.path_xml).getroot()

    def remove(self, tags):
        """remove elements by tagname"""

        for tag in tags:
            removals = self.tree.findall(f'.//{tag}', self.xmlns)
            for rem in removals:
                parent = rem.getparent()
                parent.remove(rem)
                # also clear parent textual hidden linebreaks
                if len(parent.getchildren()) == 0 and parent.text:
                    parent.text = ''

    def findall(self, expression, element=None, get_all=True):
        """wrap search by xpath-expression"""
        els = None
        if element is not None:
            els = element.findall(expression, XMLNS)
        else:
            els = self.tree.findall(expression, XMLNS)
        if els is None:
            return []
        if not get_all:
            return els[0]
        return els

    def xpath(self, expression, element=None):
        """wrap xpath 1.0 calls"""
        if element is not None:
            return element.xpath(expression, namespaces=XMLNS)
        return self.tree.xpath(expression, namespaces=XMLNS)

    def write(self, ext='.xml', out_dir=None, suffix=None) -> str:
        """Write XML-Data to METS/MODS with optional suffix"""

        file_name = self.path_xml

        # if suffix was set, trigger name creation logic below
        if suffix:
            file_name = os.path.splitext(os.path.basename(self.path_xml))[0]
            if "." in file_name:
                file_name = file_name.split(".")[0]
            if suffix:
                file_name += '.' + suffix + ext
            else:
                file_name += ext

        mets_dir = self.path_xml_dir
        if out_dir:
            if not os.path.exists(out_dir):
                raise RuntimeError(f'invalid destination {out_dir}')
            mets_dir = out_dir

        path_out = os.path.join(mets_dir, file_name)
        write_xml_file(self.tree, path_out)
        return path_out


class MetsProcessor(XMLProcessor):
    """Basic METS/MODS-Handling"""

    def enrich_agent(self, note_msg):
        """Enrich agent information"""

        mets_hdr = self.tree.find('.//mets:metsHdr', XMLNS)
        agent_smith = ET.SubElement(mets_hdr,
                                    '{http://www.loc.gov/METS/}agent',
                                    {'TYPE': 'OTHER',
                                     'ROLE': 'OTHER',
                                     'OTHERTYPE': 'SOFTWARE'})
        the_name = ET.SubElement(agent_smith, '{http://www.loc.gov/METS/}name')
        the_name.text = note_msg
        the_note = ET.SubElement(agent_smith, '{http://www.loc.gov/METS/}note')
        the_note.text = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())

    def clear_agents(self, filter_by_attr, black_list):
        """
        Drop agent entries in metsHdr section
        - usefull to drop information from intermediate OCR-Steps or Legacy tools
        - can be filtered by node@attribute with attribute blacklist
        """

        agents = self.findall('.//mets:metsHdr/mets:agent')
        for agent in agents:
            if filter_by_attr in agent.keys():
                filter_attr = agent.attrib[filter_by_attr]
                if filter_attr:
                    for black in black_list:
                        if black in filter_attr:
                            parent = agent.getparent()
                            parent.remove(agent)

    def contains_group(self, group:'str|List[str]') -> bool:
        """Test if a certain fileGroup exists"""

        if isinstance(group, list):
            return any(self.tree.findall(PATTERN_FILEGROUP_USE.format(g), XMLNS)
                       for g in group)
        return self.tree.findall(PATTERN_FILEGROUP_USE.format(group), XMLNS) > 0

    def clear_filegroups(self, black_list=None):
        """Clear file Groups by blacklist"""

        cleared = 0
        file_sections = self.tree.findall('.//mets:fileSec', XMLNS)
        if len(file_sections) < 1:
            return cleared

        sub_groups = list(file_sections[0])
        for sub_group in sub_groups:
            subgroup_label = sub_group.attrib['USE']
            if black_list:
                for black in black_list:
                    if subgroup_label == black:
                        self._clear_also_fileptrs(sub_group)
                        file_sections[0].remove(sub_group)
                        cleared += 1

        # optional: handle *now* possibly empty file section
        if len(file_sections[0].getchildren()) < 1:
            parent = file_sections[0].getparent()
            parent.remove(file_sections[0])
        return cleared

    def _clear_also_fileptrs(self, sub_group):
        """
        Drop outdated file pointer references (mets:fptr)
        not only in physical Sequence Map but also in logical structMap
        """

        files = sub_group.findall('mets:file', XMLNS)
        for _file in files:
            file_id = _file.attrib['ID']
            # self._sanitze_reference_and_page(file_id)
            references = self.tree.findall(
                f'.//mets:fptr[@FILEID="{file_id}"]', XMLNS)
            for reference in references:
                parent = reference.getparent()
                parent.remove(reference)


class MetsReaderReport:
    """Wrapper for digital object data analysis outcome"""

    def __init__(self):
        self.identifiers = {}
        self.languages = []
        self.system_identifiers = {}
        self.type = None
        self.hierarchy = None
        self.links = None
        self.images = None
        self.locations = None
        self.origins = None


PREFIX_VLS = 'md'

NEWSPAPER_TYPES = ['issue', 'additional', 'year', 'newspaper']
class MetsReader(MetsProcessor):
    """
    Read and try to make sense of METS/MODS Metadata which
    might be enclosed in OAI-PMH or as local file.
    
    Each METS-Record is considere to have at least
    * a primary descriptive Metadata section (prime_dmd), and 
    * a structure Mapping, optionally marked with @TYPE=LOGICAL
    
    For Identifiers of primary descriptive Metadata sections (prime_dmd) holds
    * VLS-Records are prefixed 'md' and the ID is considered numeric
    * Kitodo2 Monographs are fixed to "DMDLOG_0000"
    * Kitodo2 MVW are paired "DMDLOG_0000" - "DMDLOG_0001" 
    * DSpace item handle based record contain a single '/' 
    * Kitodo3 objects use UUIDs
    """

    def __init__(self, path_mets, dmd_id:str=None):
        super().__init__(path_mets)
        self._prime_type: Optional[str] = None
        self._prime_mods = None
        self._prime_mods_id: Optional[str] = dmd_id
        if self._prime_mods_id is None:
            self._determine_prime_dmd_id()
        self._determine_prime_dmd()
        self._report: Optional[MetsReaderReport] = None
        self._config = None

    @property
    def config(self):
        """Access configuration"""
        return self._config

    @config.setter
    def config(self, path_config: Path):
        """Determine configuration"""
        if not path_config.is_file():
            raise RuntimeError(f"invalid path config '{path_config}'!")
        _cp = ConfigParser()
        _cp.read(path_config)
        self._config = _cp

    def _determine_prime_dmd_id(self):
        """
        Inspect METS
        * if only single MODS present, take it
        * otherwise, guess prime by structure map
        """

        dmd_candidates = self.tree.findall(XPR_MODS_SEC, XMLNS)
        if len(dmd_candidates) == 1 and 'ID' in dmd_candidates[0].attrib:
            self._prime_mods_id = dmd_candidates[0].attrib['ID']
        else:
            # if more MODS present
            # we are only interested in those with identifiers present
            primes = [d.get('ID')
                for d in dmd_candidates
                if len([m for m in d.iterdescendants() if 'identifier' in m.tag]) > 0
            ]
            if len(primes) == 1:
                self._prime_mods_id = primes[0]
            else:
                self._prime_mods_id = self._determine_prime_dmd_id_by_logical_struct()

    def _determine_prime_dmd_id_by_logical_struct(self) -> str:
        log_struct = self.tree.find('.//mets:structMap[@TYPE="LOGICAL"]', XMLNS)
        if log_struct is None:
            raise RuntimeError(f"No logical struct in {self.tree.base}!")
        first_level = log_struct.getchildren()
        _raw_id = None
        if len(first_level) == 1:
            the_type: str = first_level[0].attrib['TYPE'].lower()
            if the_type in DEFAULT_FLAT_STRUCTS:
                _raw_id = first_level[0].attrib['DMDID']
            elif the_type in DEFAULT_PARENT_STRUCTS:
                # parent structs are considered not to have fileGroup for MAX images
                # if not self.tree.findall('.//mets:fileGrp[@USE="MAX"]', XMLNS):
                if not self.contains_group(CONTENT_FILE_GROUPS):
                    _raw_id = first_level[0].attrib['DMDID']
                else:
                    # a subsequent digital object (monograph, volume, issue) should have MAX images
                    # and further, it should posses a physical root mapping
                    _xpr_root_link = f'.//mets:structLink/mets:smLink[@{XLINK_TO}="physroot"]'
                    _root_link = self.tree.find(_xpr_root_link, XMLNS)
                    if _root_link is not None:
                        _log_id = _root_link.attrib[f'{{{XMLNS["xlink"]}}}from']
                    # no physical root link found
                    # now map most frequent xlink:from reference
                    # this must be the logical root
                    else:
                        _xpr_link = './/mets:structLink/mets:smLink'
                        _links = self.tree.findall(_xpr_link, XMLNS)
                        _link_map = defaultdict(int)
                        for _link in _links:
                            _from = _link.attrib[f'{{{XMLNS["xlink"]}}}from']
                            _link_map[_from] += 1
                        # now sort entries
                        # get first match which must be the "top-linker"
                        _log_id = sorted(_link_map, key=lambda e: e[1])[0]
                    # ... so
                    # we shall know by now the top log id
                    _xpr_log = f'.//mets:structMap[@TYPE="LOGICAL"]//mets:div[@ID="{_log_id}"]'
                    _log = self.tree.find(_xpr_log, XMLNS)
                    if _log is not None:
                        _raw_id = _log.attrib['DMDID']
        # if still no primary id found, go nuts
        if not _raw_id:
            raise RuntimeError(f"Can't find primary dmd_id in {self.tree.base}")
        return _raw_id

    def _determine_prime_dmd(self):
        """Encapsulated recognition of primary DMD section"""
        _raw_id = self._prime_mods_id
        dmd_secs = self.tree.findall(f'.//mets:dmdSec[@ID="{_raw_id}"]/mets:mdWrap/mets:xmlData/mods:mods', XMLNS)
        if len(dmd_secs) == 0:
            raise RuntimeError(f"invalid dmd_id {_raw_id}")
        # although this means invalid METS, rather check this, too
        if len(dmd_secs) > 1:
            dmd_ids = [dmd.attrib['ID'] for dmd in dmd_secs if 'ID' in dmd.attrib]
            raise RuntimeError(f"ambigious dmd_ids {dmd_ids} for {_raw_id}")
        self.primary_dmd = dmd_secs[0]

    @property
    def report(self):
        """Top level analysis"""
        if self._report is None:
            self._report = MetsReaderReport()
            self._report.identifiers = self.get_identifiers()
            self._report.system_identifiers = self._system_identifiers()
            self._report.languages = self.get_language_information()
            outcome = self.get_type_and_hierarchy()
            if outcome:
                if outcome[0]:
                    self._report.type = outcome[0]
                elif outcome[1]:
                    self._report.type = outcome[1]
                self._report.hierarchy = outcome[2]
            self._report.links = self.get_invalid_physical_structs()
            self._report.locations = self.get_location_shelfs()
            self._report.origins = self.get_origin_infos()
        return self._report

    def analyze(self):
        """Will be removed in future"""
        return self.report

    def insert_into_prime_mods(self, tag, attributes=None, text_content=None):
        """Insert element"""
        new_el = ET.SubElement(self.primary_dmd, tag, attrib=attributes)
        if text_content:
            new_el.text = text_content

    def check(self):
        """Ensure certain invariants
        
        * each logical section is at least connected to a physical container
          if not, this means empty logical sections => DDB warning, Derivans death

        Traverse only logical structure below current main logical struct 
        identified by primary descriptive section.

        raises: RuntimeError if any check fails
        """

        self._check_logical_interlinking()

    def _check_logical_interlinking(self):
        _log_ids = self.tree.xpath(f'.//mets:div[@DMDID="{self.dmd_id}"]/mets:div/@ID',
            namespaces=XMLNS)
        for _log_id in _log_ids:
            _links = self.tree.findall(f'.//mets:smLink[@{XLINK_FROM}="{_log_id}"]', XMLNS)
            if not _links:
                raise RuntimeError(f"{self.tree.base} no link for logical section:'{_log_id}'!")

    def get_type_and_hierarchy(self) -> Tuple:
        """
        Determine this item's type
        - inspect attribute TYPE of logical structMap
        - search pica entries
          https://www.dnb.de/SharedDocs/Downloads/EN/Professionell/Metadatendienste/linkedDataModellierungTiteldaten.pdf
        """

        # inspect logical structure
        (log_type, hierarchy) = self._determine_type_by_struct_logical()

        # inspect according dmd section
        pica_type = self._determine_by_dmd()

        return (pica_type, log_type, hierarchy)

    def _determine_by_dmd(self):
        """
        plain top-root .//*[@recordSyntax] will fail for multivolume single volumes because
        they might have a mods:relatedItem[@type="host"] that would also match
        """

        xp_vls_ext_type = 'mods:extension/*[@recordSyntax="pica"]'
        record_el = self.primary_dmd.find(xp_vls_ext_type, XMLNS)
        if record_el is not None:
            if 'code' in record_el.attrib:
                pica_code = record_el.attrib['code']
            else:
                pica_code = record_el.text
            return pica_code[0].upper() + pica_code[1:]
        return None

    def _determine_type_by_struct_logical(self):
        """determine logical type for source and hierarchy level"""

        xpath = f'.//*[@DMDID="{self._prime_mods_id}"]'
        log_el = self.findall(xpath, get_all=False)
        if log_el is not None:
            log_type = log_el.attrib['TYPE']
            _mods_id:str = self._prime_mods_id
            if _mods_id.startswith('md'):
                _mods_id = _mods_id[2:]
            _line = [(_mods_id, log_type)]
            level_up = log_el.getparent()
            while level_up.attrib['TYPE'] != 'LOGICAL':
                # strange issue with menadoc VL-instance
                # when OAI-URL misses context-path
                # where DMDID vanishes when stored local
                _id = 'n.a.'
                if 'DMDID' in level_up.attrib:
                    dmd_id = level_up.attrib['DMDID']
                    # somehow not elegant, but works
                    # to strip letters from legacy vk dmd:id
                    _id = dmd_id.strip('mdaslog')
                # take special care of newspaper hierarchy
                # which differs structurally
                # between mets and zmets
                # ATTENZIONE
                # take also care of Kitodo3 uuid identifiers
                # therefore this logic only applies to semantics (VLS)
                elif log_type in NEWSPAPER_TYPES and not level_up.attrib['ID'].startswith('uuid'):
                    _id = level_up.attrib['ID'].strip('log')
                else:
                    mptr_curr = level_up.findall('mets:mptr', XMLNS)
                    # exported METS from kitodo3
                    # using mptr to link parents
                    if len(mptr_curr) == 1:
                        _id = mptr_curr[0].attrib[XLINK_HREF]
                _type = level_up.attrib['TYPE']
                _line.append((_id, _type))
                level_up = level_up.getparent()
            return (log_type, _line)
        return (None, None)

    def _get_dmd_identifier(self):
        """ATTENTION!
        this will ONLY work with VL METS, when this workflows has processed
        the OAI-response data
        """
        agent_notes = self.tree.xpath('.//mets:agent/mets:note/text()', namespaces=XMLNS)
        if agent_notes:
            migration_note = [n for n in agent_notes if 'vlid' in n]
            if migration_note:
                return migration_note[0].split()[2]
        return None

    def get_identifiers(self):
        """
        Determine identifier for digital object from MODS and METS
        ATTENTION
        Requires Identifiers to be present, but this may change
        in future!
        """

        _identifiers = {**self._identifiers_from_prime_mods()}
        if not _identifiers:
            raise RuntimeError(f"no record _identifiers for {self.tree.base}!")
        if _identifiers and self._config:
            self._validate_identifier_types(_identifiers)
        return _identifiers

    def _identifiers_from_prime_mods(self) -> dict:
        """Collect identifiers and sources from prime MODS"""
        _identifiers = {}
        top_idents = self.primary_dmd.findall('mods:identifier', XMLNS)
        for top_ident in top_idents:
            _identifiers[top_ident.attrib['type']] = top_ident.text
        record_infos = self.primary_dmd.findall('mods:recordInfo/mods:recordIdentifier', XMLNS)
        for rec in record_infos:
            _identifiers[rec.attrib['source']] = rec.text
        return _identifiers

    def _system_identifiers(self) -> dict:
        """Determine system identifier(s)
        * use METS-agent-information, if available

        Please note:
        Due migration efforts there can be more
        than one system id present: actual plus 
        legacy _identifiers
        """
        _idents = {}
        # inspect mets header
        _mhdrs = self.xpath('//mets:metsHdr')
        if len(_mhdrs) == 1:
            _mhdr = _mhdrs[0]
            _repos = self.xpath('mets:agent[@OTHERTYPE="REPOSITORY"]/mets:name/text()',_mhdr)
        _repo = _repos[0] if len(_repos) == 1 else MARK_AGENT_LEGACY.split(':', maxsplit=1)[0]
        # legacy migrated record found?
        _legacy_marks = self.xpath(f'//mets:note[contains(text(), "{MARK_AGENT_LEGACY}")]/text()')
        if len(_legacy_marks) == 1:
            _id = _legacy_marks[0]
            if ':' in _id:
                _legacy_ident = _id.rsplit(':',maxsplit=1)[1].strip()
                _legacy_ident = _legacy_ident[2:] if _legacy_ident.startswith('md') else _legacy_ident
                _idents[_repo] = _legacy_ident
            else:
                _idents[_repo] = _id
        _vls_marks = self.xpath(f'//mets:note[contains(text(), "{MARK_AGENT_VLID}")]/text()')
        if len(_vls_marks) == 1:
            _id = _vls_marks[0][len(MARK_AGENT_VLID):].strip()
            _idents[_repo] = _id
        # legacy vls record which is not mapped by now?
        if _repo and ('digital' in _repo or 'menadoc' in _repo)and _repo not in _idents:
            _legacy_id = self.dmd_id[2:] if self.dmd_id.startswith('md') else self.dmd_id
            _idents[_repo] = _legacy_id
        # legacy kitodo2 source _without_ OAI envelope
        _creators = self.tree.xpath('//mets:agent[@OTHERTYPE="SOFTWARE" and @ROLE="CREATOR"]/mets:name', namespaces=XMLNS)
        if len(_creators) == 1 and 'kitodo-ugh' in _creators[0].text.lower():
            _idents[MARK_KITODO2] = None
        # kitodo3 metsDocumentID?
        _doc_ids = self.xpath('//mets:metsDocumentID/text()', _mhdr)
        if len(_doc_ids) == 1:
            _idents[MARK_KITODO3] = _doc_ids[0]
        # once migrated, now hosted at opendata
        _pres = self.tree.xpath('.//dv:presentation[contains(./text(), "://opendata")]/text()', namespaces=XMLNS)
        if len(_pres) == 1 and 'simple-search' not in _pres[0]:
            _idents[_pres[0].split('/')[2]] = _pres[0]
        if self._report and len(self._report.system_identifiers) > 0:
            _idents = {**_idents, **self._report.system_identifiers}
        if len(_idents) > 0:
            return _idents

        # we quit, no ideas left
        raise RuntimeError(f"No System _identifiers n {self.tree.base}!")

    def _validate_identifier_types(self, _identifiers: dict):
        """Transform all known _identifiers to match configuration
        regarding digital lifecycle"""

        if not self.config or not self.config.has_section('identifier'):
            return _identifiers
        _transformed = {}
        _ident_opts = self.config.options('identifier')
        for _opt in _ident_opts:
            _v = self.config.get('identifier', _opt)
            if _v in _transformed:
                raise RuntimeError(f"identifier category {_v} already in {_transformed}!")
            for k, v in _identifiers.items():
                _k = k.replace('/','.') if '/' in k else k
                if _opt.endswith('_'+_k):
                    _transformed[_v] = v
        if len(_transformed) != len(_identifiers):
            raise RuntimeError(f"Identifier transformation missmatch: {_identifiers} != {_transformed}")
        return _transformed

    def get_language_information(self):
        """Read language information"""

        xp_lang_term = 'mods:language/mods:languageTerm/text()'
        return self.primary_dmd.xpath(xp_lang_term, namespaces=XMLNS)

    def get_location_shelfs(self):
        """get data for physical prints if available; might contain multiple entries"""

        xpr_signature = f'.//mets:dmdSec[@ID="{self._prime_mods_id}"]//mods:shelfLocator/text()'
        return self.tree.xpath(xpr_signature, namespaces=XMLNS)

    def get_filegrp_links(self, group='MAX'):
        """Gather resource links for given filegroup"""

        xpath = f'.//mets:fileGrp[@USE="{group}"]/mets:file/mets:FLocat'
        resources = self.tree.findall(xpath, XMLNS)
        return [res.attrib[XLINK_HREF] for res in resources]

    def get_invalid_physical_structs(self):
        """
        Check if invalid interlinkings from logical elements
        to physical structures exist
        * keep track of preceeding valid link for later fix
        """

        xpath_links = './/mets:smLink'
        xpath_phys_pattern = './/*[@ID="{}"]'
        links = self.tree.findall(xpath_links, XMLNS)

        miss_targets = []
        latest_working = None
        for link in links:
            link_from = link.attrib[XLINK_FROM]
            link_to = link.attrib[XLINK_TO]
            xpath_phys = xpath_phys_pattern.format(link_to)
            phys_el = self.tree.find(xpath_phys, XMLNS)
            if phys_el is None:
                miss_targets.append((link_from, link_to, latest_working))
            else:
                latest_working = link_to

        return miss_targets

    def get_origin_infos(self):
        """Gather informations about origin places"""

        xp_origins = 'mods:originInfo'
        xp_place_term = 'mods:place/mods:placeTerm/text()'
        origins = self.primary_dmd.findall(xp_origins, XMLNS)

        infos = []
        for origin in origins:
            place_labels = origin.xpath(xp_place_term, namespaces=XMLNS)
            for place_label in place_labels:
                if 'eventType' in origin.attrib:
                    infos.append((origin.attrib['eventType'], place_label))
                else:
                    infos.append(('n.a.', place_label))
        return infos

    @property
    def dmd_id(self):
        """Get dmd_id as-it-is"""
        return self._prime_mods_id



# -*- coding: utf-8 -*-

import ast
import json
import os
import pytest
import shutil
import sys
import uuid
from pathlib import (
    Path
)
from unittest import (
    mock
)
from json.decoder import (
    JSONDecodeError
)

import lxml.etree as ET

from digiflow import (
    write_xml_file,
    OAIRecord,
    OAIRecordCriteriaState,
    OAIRecordCriteriaDatetime,
    OAIRecordCriteriaIdentifier,
    OAIRecordCriteriaText,
    OAIFileSweeper,
    OAIRecordHandler,
    OAILoadException,
    OAILoadServerError,
    OAILoadClientError,
    F_IDENTIFIER,
    F_SPEC,
    F_DATESTAMP,
    F_STATE_INFO,
    F_STATE,
    F_STATE_TS,
    RECORD_STATE_MASK_FRAME,
    HEADER_MIGRATION,
    OAILoader,
    post_oai_extract_mets,
    LocalStore,
    send_mail,
    request_resource,
    transform_to_record,
    get_enclosed,
)

from .conftest import (
    TEST_RES
)

ROOT = Path(__file__).parents[1]

EXPORT_METS = 'export_mets.xml'

# some test constants
ID_737429 = '737429'
OAI_ID_737429 = 'oai:digital.bibliothek.uni-halle.de/hd:{}'.format(ID_737429)
OAI_SPEC_737429 = 'ulbhaldod'
CONTENT_TYPE_TXT = 'text/xml;charset=utf-8'
OAI_BASE_URL_VD16 = 'digitale.bibliothek.uni-halle.de/vd16/oai'
OAI_BASE_URL_ZD = 'digitale.bibliothek.uni-halle.de/zd/oai'
OAI_BASE_URL_OPENDATA = 'opendata.uni-halle.de/oai/dd'


@pytest.mark.skipif("sys.version_info < (3,6)")
def test_intermediate_dirs_created_with_path(tmp_path):
    """Test depends on PosixPath, only works with 3.6+"""

    src_path = TEST_RES / "k2_mets_vd18_147638674.xml"
    path_dst = tmp_path / "sub_dir" / "another_sub_dir" / "147638674.xml"
    xml = ET.parse(src_path)

    # act
    write_xml_file(xml.getroot(), path_dst)

    assert os.path.isfile(path_dst)


def test_intermediate_dirs_created_with_tmpdir(tmpdir):
    """Test depends on PosixPath, only works with 3.6+"""

    src_path = TEST_RES / "k2_mets_vd18_147638674.xml"
    path_dst = tmpdir.join("sub_dir").mkdir().join(
        "another_sub_dir").mkdir().join("147638674.xml")
    xml = ET.parse(src_path)

    # act
    write_xml_file(xml.getroot(), str(path_dst))

    assert os.path.isfile(str(path_dst))


def test_write_xml_defaults(tmp_path):
    txt = '<parent><child name="foo">bar</child></parent>'
    xml_tree = ET.fromstring(txt)
    outpath = tmp_path / "write_foo.xml"
    write_xml_file(xml_tree, str(outpath))

    assert os.path.isfile(str(outpath))
    assert open(str(outpath)).read().startswith(
        '<?xml version="1.0" encoding="UTF-8"?>\n')


def test_write_xml_without_preamble(tmp_path):
    txt = '<parent><child name="foo">bar</child></parent>'
    xml_tree = ET.fromstring(txt)
    outpath = tmp_path / "write_foo.xml"
    write_xml_file(xml_tree, str(outpath), preamble=None)

    assert os.path.isfile(str(outpath))
    assert open(str(outpath)).read().startswith('<parent>\n')


@pytest.mark.parametrize(['urn', 'local_identifier'],
                         [
    ('oai:digital.bibliothek.uni-halle.de/hd:10595', '10595'),
    ('oai:digitale.bibliothek.uni-halle.de/vd18:9427342', '9427342'),
    ('oai:opendata.uni-halle.de:1981185920/34265', '1981185920_34265'),
    ('oai:dev.opendata.uni-halle.de:123456789/27949', '123456789_27949'),
])
def test_record_local_identifiers(urn, local_identifier):

    # act
    record = OAIRecord(urn)
    assert record.local_identifier == local_identifier


def fixture_request_results(*args, **kwargs):
    """
    Provide local copies for corresponding download request
    * dd/oai:opendata.uni-halle.de:1981185920/36020
    """
    the_url = args[0]
    the_headers = kwargs['headers'] if 'headers' in kwargs else {}
    result = mock.Mock()
    result.status_code = 200
    result.headers = {'Content-Type': 'image/jpeg'}
    if the_headers:
        for k, v in the_headers.items():
            result.headers[k] = v
        # , 'User-Agent': the_headers['User-Agent']}
    max_image_dir = os.path.join(
        str(ROOT), 'tests/resources/vls/monography/737429/MAX')
    # this one is the METS/MODS
    if the_url.endswith('36020'):
        result.headers['Content-Type'] = CONTENT_TYPE_TXT
        data_path = os.path.join(
            str(ROOT), 'tests/resources/opendata/1981185920_36020.oai.xml')
        with open(data_path, encoding="utf-8") as xml:
            result.content = xml.read()
    elif the_url.endswith('997508'):
        result.headers['Content-Type'] = CONTENT_TYPE_TXT
        data_path = os.path.join(
            str(ROOT), 'tests/resources/vls/vd16-oai-997508.xml')
        with open(data_path, encoding="utf-8") as xml:
            result.content = xml.read()
    else:
        with open(max_image_dir + '/737434.jpg', 'rb') as img:
            result.content = img.read()
    return result


@mock.patch("digiflow.requests.get")
def test_oai_load_vd16_with_localstore(mock_request_vd16_997508, tmp_path):
    """test oai loader implementation for opendata"""

    # arrange
    mock_request_vd16_997508.side_effect = fixture_request_results
    ident = 'oai:digitale.bibliothek.uni-halle.de/vd16:997508'
    record = OAIRecord(ident)
    _id = record.local_identifier
    local_dir = tmp_path / "WORKDIR" / _id
    store_dir = tmp_path / "STORE" / "dd" / _id
    local_dir.mkdir(parents=True)
    store_dir.mkdir(parents=True)
    key_images = 'MAX'
    local_dst = str(local_dir) + '/' + _id + '.xml'

    # act
    loader = OAILoader(local_dir, base_url='digitale.bibliothek.uni-halle.de/vd16/oai',
                       group_images=key_images,
                       post_oai=post_oai_extract_mets)
    loader.store = LocalStore(store_dir, local_dir)
    number = loader.load(record.identifier, local_dst, 'md997508')

    # assert first download of 1 xml + 12 image resources
    assert number == 13
    assert mock_request_vd16_997508.call_count == 13
    assert os.path.isfile(str(local_dir / (_id + ".xml")))
    assert os.path.isfile(str(local_dir / "MAX" / "1019932.jpg"))

    # ensure no subsequent re-load took place
    assert not loader.load(record.identifier, local_dst, 'md997508')
    assert mock_request_vd16_997508.call_count == 13


@mock.patch("digiflow.requests.get")
def test_oai_load_opendata_with_localstore(
        mock_request_1981185920_36020, tmp_path):
    """test oai loader implementation for opendata"""

    # arrange
    mock_request_1981185920_36020.side_effect = fixture_request_results
    ident = 'oai:opendata.uni-halle.de:1981185920/36020'
    record = OAIRecord(ident)
    _id = record.local_identifier
    local_dir = tmp_path / "WORKDIR" / _id
    store_dir = tmp_path / "STORE" / "dd" / _id
    local_dir.mkdir(parents=True)
    store_dir.mkdir(parents=True)
    key_images = 'MAX'
    local_dst = str(local_dir) + '/' + _id + '.xml'

    # act
    loader = OAILoader(local_dir, base_url=OAI_BASE_URL_OPENDATA,
                       group_images=key_images,
                       post_oai=post_oai_extract_mets)
    loader.store = LocalStore(store_dir, local_dir)
    number = loader.load(record.identifier, local_dst)

    # assert
    assert number == 12
    assert os.path.isdir(str(local_dir))
    assert os.path.isdir(str(local_dir / "MAX"))
    assert os.path.isfile(str(local_dir / (_id + ".xml")))
    assert os.path.isfile(str(local_dir / "MAX" / "00000011.jpg"))

    # check cache
    assert os.path.exists(str(store_dir))

    # check no re-load took place
    assert not loader.load(record.identifier, local_dst)


@mock.patch("digiflow.requests.get")
def test_oai_load_opendata_request_kwargs(
        mock_request_1981185920_36020, tmp_path):
    """test oai loader implementation for opendata"""

    # arrange
    mock_request_1981185920_36020.side_effect = fixture_request_results
    ident = 'oai:opendata.uni-halle.de:1981185920/36020'
    record = OAIRecord(ident)
    _id = record.local_identifier
    local_dir = tmp_path / "WORKDIR" / _id
    store_dir = tmp_path / "STORE" / "dd" / _id
    local_dir.mkdir(parents=True)
    store_dir.mkdir(parents=True)
    key_images = 'MAX'
    local_dst = str(local_dir) + '/' + _id + '.xml'
    request_kwargs = dict(headers={'User-Agent': 'Smith'})

    # act
    loader = OAILoader(local_dir, base_url=OAI_BASE_URL_OPENDATA,
                       group_images=key_images,
                       post_oai=post_oai_extract_mets,
                       request_kwargs=request_kwargs)
    loader.store = LocalStore(store_dir, local_dir)
    number = loader.load(record.identifier, local_dst)

    # assert
    assert number == 12
    assert os.path.isdir(str(local_dir))
    assert os.path.isdir(str(local_dir / "MAX"))
    assert os.path.isfile(str(local_dir / (_id + ".xml")))
    assert os.path.isfile(str(local_dir / "MAX" / "00000011.jpg"))

    # check cache
    assert os.path.exists(str(store_dir))


DEFAULT_HEADERS = "{}\t{}\t{}\t{}\t{}\t{}\n"\
    .format(F_IDENTIFIER, F_SPEC, F_DATESTAMP, F_STATE_INFO, F_STATE, F_STATE_TS)


def _write_datalist(path_data_list, data, headers=DEFAULT_HEADERS):
    with open(str(path_data_list), 'w') as handle:
        if headers:
            handle.write(headers)
        handle.writelines(data)


def test_invalid_input_data(tmp_path):
    """Invalid input data format raises an exception"""

    # arrange
    invalid_path_dir = tmp_path / 'invalid_data'
    invalid_path_dir.mkdir()
    invalid_path = invalid_path_dir / 'invalid.tsv'
    data = ["123\t456\t789\t0\n", "124\t457\t790\t1\n"]
    _write_datalist(invalid_path, data, headers=None)

    with pytest.raises(RuntimeError) as exc:
        OAIRecordHandler(invalid_path, data_fields=[F_IDENTIFIER, F_STATE])

    assert "invalid fields" in str(exc.value)


@pytest.fixture(name="valid_datasets")
def _fixture_valid_input_data(tmp_path):
    """
    provide valid input data
    * row [0] = 'IDENTIFIER' following oai-protocol
    * row [1] = 'SETSPEC' additional information on set and type (optional)
    * row [2] = 'CREATED' when item was created by legacy system
    * row [3] = 'STATE' migration state, 'n.a.' when open to migration
    * row [4] = 'FINISHED' datetime of last successfull migration run
    """
    valid_path_dir = tmp_path / 'valid_data'
    valid_path_dir.mkdir()
    valid_path = valid_path_dir / 'valid.tsv'
    data = [
        "oai:myhost.de/dod:123\tdod##book\t2009-11-03T13:20:32Z\tn.a.\tn.a.\tn.a.\n",
        "oai:myhost.de/dod:124\tdod##book\t2009-11-04T13:20:32Z\tn.a.\tn.a.\tn.a.\n"
    ]
    _write_datalist(valid_path, data)
    return str(valid_path)


def test_migrate_state_saved(valid_datasets):
    """
    Valid input data format enables to mark
    successfull outcomes by row nr.5 & nr. 6
    """

    # arrange
    handler = OAIRecordHandler(valid_datasets, mark_lock='ocr_done')
    a_set = handler.next_record()
    assert a_set.identifier == 'oai:myhost.de/dod:123'

    # act
    handler.save_record_state(a_set.identifier, state='ocr_done', INFO='444')

    # assert
    next_dataset = handler.next_record()
    assert next_dataset.identifier == 'oai:myhost.de/dod:124'

    # act
    # assert
    handler.save_record_state(next_dataset.identifier, 'ocr_done', INFO='555')

    # nothing left to do
    assert not handler.next_record()


@pytest.fixture(name="vl_datasets")
def _fixture_vl_datasets_input_data(tmp_path):
    """
    provide valid input data
    * row [0] = 'IDENTIFIER' following oai-protocol
    * row [1] = 'SETSPEC' additional information on set and type (optional)
    * row [2] = 'CREATED' when item was created by legacy system
    * row [3] = 'STATE' migration state, 'n.a.' when open to migration
    * row [4] = 'FINISHED' datetime of last successfull migration run
    """
    valid_path_dir = tmp_path / 'vl_datasets'
    valid_path_dir.mkdir()
    valid_path = valid_path_dir / 'vl_datasets.tsv'
    data = [
        'oai:menadoc.bibliothek.uni-halle.de/menalib:1416976\tmenalib\t2009-11-03T13:20:32Z\tn.a.\tn.a.\tn.a.\n',
        'oai:digitale.bibliothek.uni-halle.de/vd17:696\tpon##book\t2009-11-04T13:20:32Z\tn.a.\tn.a.\tn.a.\n'
    ]
    _write_datalist(valid_path, data)
    return str(valid_path)


def test_migration_dataset_vl_formats(vl_datasets):
    """
    Valid input data format enables to mark
    successfull outcomes by row nr.5 & nr. 6
    """

    # arrange
    handler = OAIRecordHandler(vl_datasets)
    a_set = handler.next_record()

    # assert
    assert handler.position == "0001/0002"
    assert a_set.local_identifier == '1416976'
    assert a_set.set == 'menalib'
    handler.save_record_state(a_set.identifier, 'busy')

    a_set = handler.next_record()
    assert handler.position == "0002/0002"
    assert a_set.local_identifier == '696'
    assert a_set.set == 'pon##book'


def test_migration_dataset_vl_info_stays(vl_datasets):
    """
    Valid input data format enables to mark
    successfull outcomes by row nr.5 & nr. 6
    """

    # arrange
    handler = OAIRecordHandler(vl_datasets)
    a_set = handler.next_record()

    # assert
    assert a_set.local_identifier == '1416976'
    handler.save_record_state(a_set.identifier, 'metadata_done', INFO='123,ger')
    handler.save_record_state(a_set.identifier, 'migration_done')

    a_set = handler.next_record(state='migration_done')
    assert a_set.local_identifier == '1416976'
    with open(vl_datasets) as reader:
        lines = reader.readlines()
    assert lines[1].split('\t')[3] == '123,ger'


def test_dataset_cannot_find_entry(vl_datasets):
    """Behavior if want to store with unknown identifier"""

    # arrange
    handler = OAIRecordHandler(vl_datasets)

    # act
    with pytest.raises(RuntimeError) as exc:
        handler.save_record_state('foo')

    # assert
    assert 'No Record for foo' in str(exc.value)


def test_statelist_ocr_hdz(tmp_path):
    """Behavior of state lists for ocr pipeline"""

    # arrange
    path_state_list = tmp_path / 'ocr_list'
    first_row = "{}\t{}\t{}\n".format(F_IDENTIFIER, F_STATE, F_STATE_TS)
    data = [
        'oai:digitale.bibliothek.uni-halle.de/zd:123\tn.a.\tn.a.\n'
    ]
    _write_datalist(path_state_list, data, first_row)
    headers = [F_IDENTIFIER, F_STATE, F_STATE_TS]
    handler = OAIRecordHandler(
        path_state_list,
        data_fields=headers,
        transform_func=lambda r: r)

    # act
    record = handler.next_record()
    assert handler.position == '0001/0001'
    assert record[F_IDENTIFIER] == 'oai:digitale.bibliothek.uni-halle.de/zd:123'
    handler.save_record_state(record[F_IDENTIFIER])

    # assert no next open record
    assert not handler.next_record()

    # assert there is a record next with default state 'lock'
    assert handler.next_record('busy')


# record list header data
OAI_LIST_HEADER = "{}\t{}\t{}\t{}\t{}\n".format(
    F_IDENTIFIER, F_DATESTAMP, F_STATE_INFO, F_STATE, F_STATE_TS)


@pytest.fixture(name="oai_record_list")
def _fixture_oai_record_list(tmp_path):
    path_state_list = tmp_path / 'ocr_list'
    data = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853011\t2015-08-25T20:00:35Z\tn.a.\tocr_skip\t2021-08-03_15:03:56\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:17320046\t2021-09-01T15:25:43Z\t17320046,issue,ger,20\tupload_done\t2021-09-09_22:57:45\n"
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tn.a.\tocr_skip\t2021-08-03_15:14:45\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:8853013\t2015-08-25T20:00:35Z\tn.a.\tocr_skip\t2021-08-03_15:20:45\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:9510507\t2015-08-25T20:00:35Z\t9510507,issue,[('9510507', 'issue'), ('n.a.', 'section'), ('8849883', 'year'), ('9059307', 'newspaper')],,None\tocr_done\t2021-08-03_16:44:54\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:9510508\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n"
    ]
    _write_datalist(path_state_list, data, OAI_LIST_HEADER)
    return path_state_list


def _morph(row):
    r = {}
    r[F_IDENTIFIER] = row[F_IDENTIFIER]
    r[F_STATE] = row[F_STATE]
    r[F_STATE_TS] = row[F_STATE_TS]
    return r


# OAI record list default headers
headers = [F_IDENTIFIER, F_DATESTAMP, F_STATE_INFO, F_STATE, F_STATE_TS]


def test_statelist_use_n_a_properly(oai_record_list):
    """Behavior of state lists for ocr pipeline"""

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        data_fields=headers,
        transform_func=_morph)

    # act
    record = handler.next_record()
    assert handler.position == '0006/0006'
    assert record[F_IDENTIFIER] == 'oai:digitale.bibliothek.uni-halle.de/zd:9510508'
    handler.save_record_state(record[F_IDENTIFIER])

    # assert no next open record
    assert not handler.next_record()

    # assert there is a record with default state 'lock'
    assert handler.next_record('busy')


def fixture_request_vls_zd1_16359609(*args, **kwargs):
    """
    Provide local copies for corresponding download request
    Data: oai:digitale.bibliothek.uni-halle.de/zd:16359609
    """
    the_url = args[0]
    the_headers = kwargs['headers'] if 'headers' in kwargs else {}
    result = mock.Mock()
    result.status_code = 200
    result.headers = {'Content-Type': CONTENT_TYPE_TXT}
    if the_headers:
        for k, v in the_headers.items():
            result.headers[k] = v
        # , 'User-Agent': the_headers['User-Agent']}
    max_image_dir = os.path.join(
        str(ROOT), 'tests/resources/vls/monography/737429/MAX')
    # this one is the METS/MODS
    if the_url.endswith('16359609'):
        data_path = os.path.join(
            str(ROOT), 'tests/resources/vls/zd/zd1-16359609.mets.xml')
        with open(data_path, encoding="utf-8") as xml:
            result.content = xml.read()
    elif 'download/webcache/' in the_url:
        result.headers = {'Content-Type': 'image/jpeg'}
        with open(max_image_dir + '/737434.jpg', 'rb') as img:
            result.content = img.read()
    elif 'download/fulltext/' in the_url:
        alto_file = os.path.join(str(ROOT), 'tests/resources/vls/zd/')
        with open(alto_file + 'zd1-alto-16331001.xml', encoding="utf-8") as hndl:
            result.content = hndl.read().encode()
    return result


@mock.patch("digiflow.requests.get")
def test_oai_load_vls_zd1_with_ocr(mock_request, tmp_path):
    """Behavior of state lists for ocr pipeline"""

    # arrange
    """test oai loader implementation for opendata"""

    # arrange
    mock_request.side_effect = fixture_request_vls_zd1_16359609
    ident = 'oai:digitale.bibliothek.uni-halle.de/zd:16359609'
    record = OAIRecord(ident)
    _id = record.local_identifier
    local_dir = tmp_path / "WORKDIR" / _id
    store_dir = tmp_path / "STORE" / "zd" / _id
    local_dir.mkdir(parents=True)
    store_dir.mkdir(parents=True)
    local_dst = str(local_dir) + '/' + _id + '.xml'

    def post_write_mets(the_self, the_data):
        """Just extract METS from OAI body"""
        xml_root = ET.fromstring(the_data)
        write_xml_file(xml_root, the_self.path_mets, preamble=None)
        return the_self.path_mets

    # act
    loader = OAILoader(local_dir, base_url=OAI_BASE_URL_ZD,
                       post_oai=post_write_mets)
    loader.store = LocalStore(store_dir, local_dir)
    number = loader.load(record.identifier, local_dst)

    # assert
    assert number == 17
    assert os.path.isdir(str(local_dir))
    assert os.path.isdir(str(local_dir / "MAX"))
    assert os.path.isfile(str(local_dir / (_id + ".xml")))
    assert os.path.isfile(str(local_dir / "MAX" / "16331052.jpg"))

    # check cache
    assert os.path.exists(str(store_dir))


@pytest.fixture(name="migration_sweeper_img_fixture")
def _fixture_migration_img_sweeper(tmp_path):
    """Provide test fixture"""

    testroot = tmp_path / 'SWEEPERTEST'
    testroot.mkdir()
    maxdir = testroot / 'MAX'
    file_name = '737429.mets.xml'
    testresources = Path(ROOT / 'tests' / 'resources' / 'migration')
    mets_source = testresources / file_name
    shutil.copy(str(mets_source), os.path.join(str(testroot), file_name))
    images_dir = testresources / 'MAX'
    shutil.copytree(str(images_dir), str(maxdir))
    return str(testroot)


def test_migration_sweeper_img(migration_sweeper_img_fixture):
    """Test cleanup images and renaming of mets file"""

    OAIFileSweeper(migration_sweeper_img_fixture).sweep()

    for item in Path(migration_sweeper_img_fixture).iterdir():
        if str(item.name) == 'MAX':
            # preserve colorchecker
            assert len(list(item.iterdir())) == 1


@pytest.fixture(name="migration_sweeper_pdf_fixture")
def _fixture_migration_pdf_sweeper(tmp_path):
    """Provide test fixture"""

    testroot = tmp_path / 'SWEEPERTEST'
    testroot.mkdir()
    downloaddir = testroot / 'DOWNLOAD'
    downloaddir.mkdir()
    file_name = '1981185920_44046.xml'
    pdf_file_name = '265982944.pdf'
    testresources = Path(ROOT / 'tests' / 'resources' / 'ocr')
    mets_source = testresources / file_name
    shutil.copy(str(mets_source), os.path.join(str(testroot), file_name))
    #  downloaddir.write_bytes('data')
    with open(downloaddir / pdf_file_name, 'wb') as fh:
        fh.write(b'arbitrary pdf')
    return str(testroot)


def test_migration_sweeper_pdf(migration_sweeper_pdf_fixture):
    """Test cleanup images and renaming of mets file"""

    oais = OAIFileSweeper(
        migration_sweeper_pdf_fixture, pattern=".xml", filegroups=['DOWNLOAD'])
    oais.sweep()

    for item in Path(migration_sweeper_pdf_fixture).iterdir():
        if str(item.name) == 'DOWNLOAD':
            assert len(list(item.iterdir())) == 0


@mock.patch('digiflow.digiflow_io.get_smtp_server')
def test_send_mail(mock_smtp):
    """test sending mail"""

    # arrange
    random_message = uuid.uuid4().hex

    # act
    mess = send_mail(
        subject='test',
        message=random_message,
        sender='test@example.com',
        recipients='me@example.de')

    # assert
    assert random_message in mess
    assert 'notification' in mess
    assert mock_smtp.called


def test_record_state_list_set_state_from(oai_record_list):
    """Behavior for changing state matching state and start time"""

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        data_fields=headers,
        transform_func=_morph)
    c_state = OAIRecordCriteriaState('ocr_skip')
    c_from = OAIRecordCriteriaDatetime(dt_from='2021-08-03_15:03:56')

    # pre-check
    record1 = handler.next_record()
    assert record1[F_IDENTIFIER].endswith('9510508')

    # act
    outcome = handler.states(criterias=[c_state, c_from], dry_run=False)

    # assert next open record changed
    record2 = handler.next_record()
    assert record2[F_IDENTIFIER].endswith('8853011')
    # assert we handled 3 records
    assert outcome == 3


def test_record_datestamp(oai_record_list):
    """Check if proper datestamp gets picked"""

    # arrange
    _handler = OAIRecordHandler(oai_record_list)

    # act
    _record = _handler.next_record()

    # assert
    assert _record.identifier == 'oai:digitale.bibliothek.uni-halle.de/zd:9510508'
    assert _record.local_identifier == '9510508'
    assert _record.date_stamp == '2015-08-25T20:00:35Z'


def test_record_get_fullident(oai_record_list):
    """Check if proper datestamp gets picked"""

    # arrange
    _ident_urn = 'oai:digitale.bibliothek.uni-halle.de/zd:9510508'
    _handler = OAIRecordHandler(oai_record_list)

    # act
    _record = _handler.get(_ident_urn)

    # assert
    assert _record.identifier == _ident_urn
    assert _record.date_stamp == '2015-08-25T20:00:35Z'


def test_record_get_partialident(oai_record_list):
    """Check if proper datestamp gets picked
    if just the latest segment (aka local
    identifier) has been provided"""

    # arrange
    _ident_urn = 'oai:digitale.bibliothek.uni-halle.de/zd:9510508'
    _handler = OAIRecordHandler(oai_record_list)

    # act
    _record_exact = _handler.get('9510508')
    _record_fuzzy = _handler.get('9510508', exact_match=False)

    # assert
    assert not _record_exact
    assert _record_fuzzy.identifier == _ident_urn
    assert _record_fuzzy.date_stamp == '2015-08-25T20:00:35Z'


def test_record_get_non_existent(oai_record_list):
    """What happens if no match found?"""

    # arrange
    _handler = OAIRecordHandler(oai_record_list)

    # act
    assert not _handler.get('9510509')


def test_record_state_list_set_state_from_dry_run(oai_record_list):
    """Behavior of dry run"""

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        data_fields=headers,
        transform_func=_morph)
    c_state = OAIRecordCriteriaState('ocr_skip')
    c_from = OAIRecordCriteriaDatetime(dt_from='2021-08-03_15:03:56')

    # pre-check
    record1 = handler.next_record()
    assert record1[F_IDENTIFIER].endswith('9510508')

    # act
    outcome = handler.states(criterias=[c_state, c_from], dry_run=True)

    # assert no change of date
    record2 = handler.next_record()
    assert record2[F_IDENTIFIER].endswith('9510508')
    assert outcome == 3


def test_record_state_list_set_state_from_dry_run_verbose(oai_record_list, capsys):
    """Behavior of dry run"""

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        data_fields=headers,
        transform_func=_morph)
    c_state = OAIRecordCriteriaState('ocr_skip')
    c_from = OAIRecordCriteriaDatetime(dt_from='2021-08-03_15:03:56')

    # pre-check
    record1 = handler.next_record()
    assert record1[F_IDENTIFIER].endswith('9510508')

    # act
    outcome = handler.states(criterias=[c_state, c_from], dry_run=True, verbose=True)

    # assert
    assert outcome == 3
    _std_out = capsys.readouterr().out
    assert 'IDENTIFIER\tCREATED\tINFO\tSTATE\tSTATE_TIME' in _std_out
    assert 'oai:digitale.bibliothek.uni-halle.de/zd:8853012' in _std_out


def test_record_state_list_rewind_state_upload(oai_record_list):
    """Behavior for changing state"""

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        data_fields=headers,
        transform_func=_morph)
    c_state = OAIRecordCriteriaState('upload_done')

    # pre-check
    record1 = handler.next_record()
    assert record1[F_IDENTIFIER].endswith('9510508')

    # act
    handler.states(criterias=[c_state], dry_run=False)
    assert handler.next_record()[F_IDENTIFIER].endswith('17320046')

    # assert next open record changed
    h2 = OAIRecordHandler(
        oai_record_list, data_fields=headers, transform_func=_morph)
    assert h2.next_record()[F_IDENTIFIER].endswith('17320046')


def test_record_set_some_other_state(oai_record_list):
    """Behavior for changing state"""

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        data_fields=headers,
        transform_func=_morph)
    c_state = OAIRecordCriteriaState('upload_done')

    # pre-check
    record1 = handler.next_record()
    assert record1[F_IDENTIFIER].endswith('9510508')

    # act
    handler.states(
        criterias=[c_state], set_state='metadata_read', dry_run=False)
    assert not handler.next_record(state='upload_done')

    # assert next open record changed
    h2 = OAIRecordHandler(
        oai_record_list, data_fields=headers, transform_func=_morph)
    assert h2.next_record(state='metadata_read')[F_IDENTIFIER].endswith('17320046')


def test_record_list_time_range(oai_record_list):
    """Behavior for setting state within time range"""

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        data_fields=headers,
        transform_func=_morph)
    c_state = OAIRecordCriteriaState('ocr_skip')
    c_range = OAIRecordCriteriaDatetime(dt_from='2021-08-03_15:10:00',
                                        dt_to='2021-08-03_15:20:00')

    # pre-check
    record1 = handler.next_record()
    assert record1[F_IDENTIFIER].endswith('9510508')

    # act
    outcome = handler.states(criterias=[c_state, c_range], dry_run=False)
    assert outcome == 1

    # assert next open record changed
    record2 = handler.next_record()
    assert record2[F_IDENTIFIER].endswith('8853012')


def test_record_handler_merge_plain(tmp_path):
    """Behavior if merging 2 lists"""

    # arrange
    path_oai_list1 = tmp_path / 'oai_list1'
    data1 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:8853013\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:9510507\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n"
    ]
    _write_datalist(path_oai_list1, data1, OAI_LIST_HEADER)
    path_oai_list2 = tmp_path / 'oai_list2'
    data2 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tinfo1\tupload_done\t2021-08-03_15:14:45\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:9510507\t2015-08-25T20:00:35Z\tinfo2\tocr_fail\t2021-08-03_16:14:45\n"
    ]
    _write_datalist(path_oai_list2, data2, OAI_LIST_HEADER)
    handler = OAIRecordHandler(
        path_oai_list1,
        data_fields=headers,
        transform_func=_morph)

    # assert original next == first record with identifier 8853012
    assert handler.next_record()[F_IDENTIFIER].endswith('8853012')

    # act
    merge_result = handler.merges(path_oai_list2, dry_run=False)

    # list1 and list2 contain same records
    assert merge_result['matches'] == 2
    # list2 didn't contain any records *not* included in list1
    assert merge_result['misses'] == 0
    # two records have been merged
    assert merge_result['merges'] == 2
    # no new records
    assert merge_result['appendeds'] == 0
    # next must have changed, since 'n.a.' was merged with 'upload_done'
    assert handler.next_record()[F_IDENTIFIER].endswith('8853013')
    # state from record with ident '8853012' is now 'upload_done'
    criterias = [OAIRecordCriteriaIdentifier(
        '8853012'), OAIRecordCriteriaState('upload_done')]
    assert handler.states(criterias=criterias) == 1


def test_record_handler_merge_only_done(tmp_path):
    """Behavior if merging 2 lists with specific requirement"""

    # arrange
    path_oai_list1 = tmp_path / 'oai_list1'
    data1 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:8853013\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:9510507\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n"
    ]
    _write_datalist(path_oai_list1, data1, OAI_LIST_HEADER)
    path_oai_list2 = tmp_path / 'oai_list2'
    data2 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tinfo1\tupload_done\t2021-08-03_15:14:45\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:9510507\t2015-08-25T20:00:35Z\tinfo2\tocr_fail\t2021-08-03_16:14:45\n"
    ]
    _write_datalist(path_oai_list2, data2, OAI_LIST_HEADER)
    handler = OAIRecordHandler(
        path_oai_list1,
        data_fields=headers,
        transform_func=_morph)

    # act
    merge_result = handler.merges(path_oai_list2, dry_run=False, other_require_state='upload_done')

    # list1 and list2 contain same records
    assert merge_result['matches'] == 2
    # list2 didn't contain any records *not* included in list1
    assert merge_result['misses'] == 0
    # only the required record was merged
    assert merge_result['merges'] == 1
    # no new records
    assert merge_result['appendeds'] == 0
    # next must have changed, since 'n.a.' was merged with 'upload_done'
    assert handler.next_record()[F_IDENTIFIER].endswith('8853013')
    # state from record with ident '8853012' is now 'upload_done'
    criterias = [OAIRecordCriteriaIdentifier(
        '8853012'), OAIRecordCriteriaState('upload_done')]
    assert handler.states(criterias=criterias) == 1


def test_record_handler_merge_ignore_failure(tmp_path):
    """Behavior if merging 2 lists but ignore failures"""

    # arrange
    path_oai_list1 = tmp_path / 'oai_list1'
    data1 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:8853013\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:9510507\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n"
    ]
    _write_datalist(path_oai_list1, data1, OAI_LIST_HEADER)
    path_oai_list2 = tmp_path / 'oai_list2'
    data2 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tinfo1\tupload_done\t2021-08-03_15:14:45\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:9510507\t2015-08-25T20:00:35Z\tinfo2\tocr_fail\t2021-08-03_16:14:45\n"
    ]
    _write_datalist(path_oai_list2, data2, OAI_LIST_HEADER)
    handler = OAIRecordHandler(
        path_oai_list1,
        data_fields=headers,
        transform_func=_morph)

    # act
    merge_result = handler.merges(path_oai_list2, dry_run=False, other_ignore_state='upload_done')

    # list1 and list2 contain same records
    assert merge_result['matches'] == 2
    # list2 didn't contain records *not* included in list1
    assert merge_result['misses'] == 0
    # only the required record was merged
    assert merge_result['merges'] == 1
    # no new records
    assert merge_result['appendeds'] == 0
    # next is still first record
    assert handler.next_record()[F_IDENTIFIER].endswith('8853012')
    # state from record with ident '9510507' is now 'ocr_fail'
    criterias = [OAIRecordCriteriaIdentifier(
        '9510507'), OAIRecordCriteriaState('ocr_fail')]
    assert handler.states(criterias=criterias) == 1


def test_record_handler_merge_larger_into_smaller_dry_run(tmp_path):
    """Behavior if larger list gets merged into smaller.
    Question: How to deal with unknown dataset?
    """

    # arrange
    path_oai_list_a = tmp_path / 'oai_list_a'
    data = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tinfo1\tupload_done\t2021-08-03_15:14:45\n"
    ]
    _write_datalist(path_oai_list_a, data, OAI_LIST_HEADER)
    handler = OAIRecordHandler(
        path_oai_list_a,
        data_fields=headers,
        transform_func=_morph)
    path_oai_list_b = tmp_path / 'oai_list_b'
    data = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:8853013\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n"
    ]
    _write_datalist(path_oai_list_b, data, OAI_LIST_HEADER)

    # assert original next == first record with identifier 8853012
    assert handler.next_record(state='upload_done')[
        F_IDENTIFIER].endswith('8853012')

    # act: must set ignore_state=None,
    # otherwise other list's records will be
    # completely ignored
    result = handler.merges(path_oai_list_b, other_ignore_state=None)

    # first record from self saved as-it-was, no merge
    assert result['merges'] == 0
    # second record from other not found in self, therefore missed
    assert result['misses'] == 1
    # second record from other also finally appended
    assert result['appendeds'] == 1

    # we still expected only '1' records to be in handler list a
    assert handler.total_len == 1


def test_record_handler_merge_larger_into_smaller_hot_run_subsequent(tmp_path):
    """Two lists merged into one list with two records.
    *Originally* this did only work because
    variable 'self_record' was *already* initialized by
    first run - although resulted into inconsistent data
    with resulting list containing 2x record no.1!"""

    # arrange
    path_oai_list_a = tmp_path / 'oai_list_a'
    data1 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tinfo1\tupload_done\t2021-08-03_15:14:45\n"
    ]
    _write_datalist(path_oai_list_a, data1, OAI_LIST_HEADER)
    handler = OAIRecordHandler(
        path_oai_list_a,
        data_fields=headers,
        transform_func=_morph)
    path_oai_list_b = tmp_path / 'oai_list_b'
    data2 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:8853013\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n"
    ]
    _write_datalist(path_oai_list_b, data2, OAI_LIST_HEADER)

    # act: must *not* set ignore_state=None,
    # otherwise first record from second list erases
    # existing data from list a
    results = handler.merges(path_oai_list_b, dry_run=False)

    # no merge since state 'n.a.' from list b
    # first record of list b ignored
    # to preserve existing data from list a
    assert results['merges'] == 0
    assert results['ignores'] == 1
    assert results['appendeds'] == 1
    # now we expected '2' records to be in handler list
    assert handler.total_len == 2
    # now this is the first result record ...
    assert handler.next_record(state='upload_done')[F_IDENTIFIER].endswith('8853012')
    # ... and this shall be second record by now
    assert handler.next_record()[F_IDENTIFIER].endswith('8853013')


def test_record_handler_merge_larger_into_smaller_hot_run_inverse(tmp_path):
    """Ensure following Error is gone:
    'UnboundLocalError: local variable 'self_record' referenced before assignment'

    Yielded because record 1 from list 2 is unknown.
    """

    # arrange
    path_oai_list_a = tmp_path / 'oai_list_a'
    data1 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tinfo1\tupload_done\t2021-08-03_15:14:45\n"
    ]
    _write_datalist(path_oai_list_a, data1, OAI_LIST_HEADER)
    handler = OAIRecordHandler(
        path_oai_list_a,
        data_fields=headers,
        transform_func=_morph)
    path_oai_list_b = tmp_path / 'oai_list_b'
    data2 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853013\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n"
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
    ]
    _write_datalist(path_oai_list_b, data2, OAI_LIST_HEADER)

    # act:
    # ignore_state=None to overwrite
    # record 1 from list a (=merge)
    results = handler.merges(path_oai_list_b, dry_run=False, other_ignore_state=None)
    assert results['merges'] == 1
    assert results['ignores'] == 0
    assert results['appendeds'] == 1

    # two records have been merged
    # now we expected '2' records to be in handler list
    assert handler.total_len == 2
    # now this is the first result record ...
    record_012 = handler.next_record()
    assert record_012[F_IDENTIFIER].endswith('8853012')
    # mark state record 1
    handler.save_record_state(record_012[F_IDENTIFIER], state='foo_bar')
    # ... this shall be second record by now
    assert handler.next_record()[F_IDENTIFIER].endswith('8853013')


def test_record_handler_merge_cross(tmp_path):
    """
    Behavior merging 2 lists with cross-different
    info, with info from list 1 must be preserved
    but new info from list 2 must be integrated
    """

    # arrange
    path_oai_list1 = tmp_path / 'oai_list1'
    data1 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tinfo1\tupload_done\t2021-08-03_15:14:45\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:8853013\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
    ]
    _write_datalist(path_oai_list1, data1, OAI_LIST_HEADER)
    path_oai_list2 = tmp_path / 'oai_list2'
    data2 = [
        "oai:digitale.bibliothek.uni-halle.de/zd:8853012\t2015-08-25T20:00:35Z\tn.a.\tn.a.\tn.a.\n",
        "oai:digitale.bibliothek.uni-halle.de/zd:8853013\t2015-08-25T20:00:35Z\tinfo2\tocr_fail\t2021-08-03_16:14:45\n"
    ]
    _write_datalist(path_oai_list2, data2, OAI_LIST_HEADER)
    handler = OAIRecordHandler(
        path_oai_list1,
        data_fields=headers,
        transform_func=_morph)

    # assert original next == first record with identifier 8853012
    assert handler.next_record()[F_IDENTIFIER].endswith('8853013')

    # act
    results = handler.merges(path_oai_list2, dry_run=False)

    # act
    assert results['merges'] == 1  # because info got merged from 2 record
    assert results['ignores'] == 1  # nothing missed, because all idents present
    assert not results['appendeds']  # no new, because all idents present
    assert not handler.next_record(state='n.a.')    # by now no more open records


def test_records_frame_with_start(oai_record_list):

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        data_fields=headers,
        transform_func=_morph)

    # act
    new_path = handler.frame(2)

    # assert
    assert os.path.exists(new_path)
    assert os.path.basename(new_path) == 'ocr_list_02_06.csv'

    frame_handler = OAIRecordHandler(
        new_path,
        data_fields=headers,
        transform_func=_morph)

    # ensure that only first records is set to 'other_load'
    # all other records are within frame
    c_state = OAIRecordCriteriaState('other_load')
    assert frame_handler.states([c_state]) == 1


def test_records_frame_range(oai_record_list):

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        data_fields=headers,
        transform_func=_morph)

    # act
    new_path = handler.frame(2, 2)

    # assert
    assert os.path.exists(new_path)
    assert os.path.basename(new_path) == 'ocr_list_02_03.csv'

    frame_handler = OAIRecordHandler(
        new_path,
        data_fields=headers,
        transform_func=_morph)

    # ensure that by now 4 records are set to 'other_load'
    # first record and records 4,5,6
    c_state = OAIRecordCriteriaState('other_load')
    assert frame_handler.states([c_state]) == 4


def test_records_default_header_from_file(oai_record_list):

    # arrange
    handler = OAIRecordHandler(
        oai_record_list,
        transform_func=_morph)

    # act
    new_path = handler.frame(3)

    # assert
    assert os.path.exists(new_path)
    assert os.path.basename(new_path) == 'ocr_list_03_06.csv'

    frame_handler = OAIRecordHandler(
        new_path,
        transform_func=_morph)

    # ensure that by now 4 records are set to 'other_load'
    # first + second record only
    c_state = OAIRecordCriteriaState(RECORD_STATE_MASK_FRAME)
    assert frame_handler.states([c_state]) == 2


def mock_response(**kwargs):
    """Create custum mock object"""

    _response = mock.MagicMock()
    _response.reason = 'testing reason'
    if 'reason' in kwargs:
        _response.reason = kwargs['reason']
    if 'status_code' in kwargs:
        _response.status_code = int(kwargs['status_code'])
    if 'headers' in kwargs:
        _response.headers = kwargs['headers']
    if 'data_path' in kwargs:
        with open(kwargs['data_path'], encoding="utf-8") as xml:
            _response.content = xml.read().encode()
    return _response


@mock.patch('requests.get')
def test_response_404(mock_requests):
    """test request ends up with 417"""

    # arrange
    _req = mock_response(status_code=417)
    mock_requests.return_value = _req

    # act
    with pytest.raises(OAILoadException) as exc:
        request_resource('http://foo.bar', Path())

    # assert
    assert exc.typename == 'OAILoadClientError'
    assert "url 'http://foo.bar' returned '417'" == exc.value.message


@mock.patch('requests.get')
def test_response_200_with_error_content(mock_requests):
    """test request results into OAILoadException"""

    # arrange
    data_path = os.path.join(str(ROOT), 'tests/resources/opendata/id_not_exist.xml')
    _req = mock_response(status_code=200, 
                         headers= {'Content-Type': 'text/xml;charset=UTF-8'},
                         data_path=data_path)
    mock_requests.return_value = _req

    # act
    with pytest.raises(OAILoadException) as exc:
        request_resource('http://foo.bar', Path())

    assert 'verb requires' in str(exc.value)


def test_records_sample_zd1_post_ocr():

    path_list = os.path.join(str(ROOT), 'tests', 'resources',
                             'vls', 'oai-urn-zd1-sample70k.tsv')
    assert os.path.isfile(path_list)

    # arrange
    handler = OAIRecordHandler(
        path_list)
    crit1 = OAIRecordCriteriaDatetime(dt_from='2021-10-16_09:45:00')

    # assert
    assert 1 == handler.states([crit1])
    crit2 = OAIRecordCriteriaState('other_load')
    assert 10 == handler.states([crit2])


def test_recordcriteria_with_created_datetime_format():
    """Expect 12 records to have a CREATED datetime before 15:26:00"""

    path_list = os.path.join(str(ROOT), 'tests', 'resources',
                             'vls', 'oai-urn-zd1-sample70k.tsv')
    assert os.path.isfile(path_list)

    # arrange
    handler = OAIRecordHandler(
        path_list)
    crit1 = OAIRecordCriteriaDatetime(
        dt_field='CREATED',
        dt_to='2021-09-01T15:26:00Z',
        dt_format='%Y-%m-%dT%H:%M:%SZ')

    # assert
    assert 12 == handler.states([crit1])


def test_record_handler_search_info(tmp_path):
    """
    Behavior searching special textual content in a specific column
    """

    # arrange
    path_oai_list1 = tmp_path / 'oai-list'
    the_data = [
        ("oai:digitale.bibliothek.uni-halle.de/vd18:1178220\tulbhalvd18##book\t2009-11-23T10:51:32Z\t"
         "683567713,Aa,vd18#10198547/urn#urn:nbn:de:gbv:3:1-114513,[],lat,['AB 71 B 3/g, 23'],582 errs:{'no_publ_place': '683567713'},no colorchecker\t"
         "fail\t2021-12-08_13:08:14\n"),
        ("oai:digitale.bibliothek.uni-halle.de/vd18:1177464\tulbhalvd18##book\t2009-11-24T07:23:00Z\t"
         "30959913X,Aa,vd18#1009007X/gbv#30959913X/urn#urn:nbn:de:gbv:3:1-114858,[],ger,['AB 95878'],472 errs:no colorchecker\t"
         "fail\t2021-12-08_13:10:52\n"),
        ("oai:digitale.bibliothek.uni-halle.de/vd18:1178423\tulbhalvd18##book\t2009-11-18T08:39:21Z\t"
         "242994199,Aa,vd18#10084061/gbv#242994199/urn#urn:nbn:de:gbv:3:1-114422,[],fre#ger,['AB 39 12/k, 21'],479,cc\t"
         "migration_done\t2021-12-08_12:36:15\n")
    ]
    _write_datalist(path_oai_list1, the_data)
    handler = OAIRecordHandler(
        path_oai_list1,
        data_fields=HEADER_MIGRATION,
        transform_func=_morph)
    crit1 = OAIRecordCriteriaText('no colorchecker')
    crit2 = OAIRecordCriteriaText('no_publ_place')

    # assert original next == nothing open
    assert not handler.next_record()
    # records contain token "no colorchecker" in their INFO ...
    assert 2 == handler.states([crit1])
    # ... but only one record matches both
    assert 1 == handler.states([crit1, crit2])

    # act
    affected = handler.states([crit1, crit2], dry_run=False)
    assert affected == 1
    assert handler.next_record()    # by now one open record


def test_record_handler_quotation_broken(tmp_path):
    """
    Fix behavior when encountering data with
    broken format from a mixture of single and double quotes

    originates from ODEM project csv list
    """

    # arrange
    path_list_sample = os.path.join(TEST_RES, 'oai-records-opendata-vd18-sample-broken.csv')
    path_oai_list1 = tmp_path / 'oai-list.csv'
    shutil.copy(path_list_sample, path_oai_list1)
    handler = OAIRecordHandler(path_oai_list1, transform_func=transform_to_record)

    # act
    _next_rec = handler.next_record(state='ocr_fail')    
    assert _next_rec.info.startswith('141.48.10.202@2023-01-17_15:55:49')
    with pytest.raises(JSONDecodeError) as _decode_err:
        json.loads(_next_rec.info)
    assert 'Extra data: line 1 column 7 (char 6)' in _decode_err.value.args[0]


def test_record_handler_quotation_mixture_json(tmp_path):
    """
    Fix behavior when encountering data which
    has been fixed manually resulting in a mixture
    of double and single quotes

    originates from ODEM project csv list
    """

    # arrange
    path_list_sample = os.path.join(TEST_RES, 'oai-records-opendata-vd18-sample-mixture.csv')
    path_oai_list1 = tmp_path / 'oai-list.csv'
    shutil.copy(path_list_sample, path_oai_list1)
    handler = OAIRecordHandler(path_oai_list1, transform_func=transform_to_record)

    # act
    _next_rec = handler.next_record(state='ocr_fail')    
    assert _next_rec.info.startswith('141.48.10.202@2023-01-17_15:55:49')
    with pytest.raises(JSONDecodeError) as _decode_err:
        json.loads(_next_rec.info)
    assert 'Extra data: line 1 column 7 (char 6)' in _decode_err.value.args[0]

def test_record_handler_quotation_fixed_json(tmp_path):
    """
    Fix behavior when encountering data which
    has been fixed manually resulting in a mixture
    of double and single quotes

    originates from ODEM project csv list
    """

    # arrange
    path_list_sample = os.path.join(TEST_RES, 'oai-records-opendata-vd18-sample-fixed.csv')
    path_oai_list1 = tmp_path / 'oai-list.csv'
    shutil.copy(path_list_sample, path_oai_list1)
    handler = OAIRecordHandler(path_oai_list1, transform_func=transform_to_record)

    # act
    _next_rec = handler.next_record(state='ocr_fail')    
    assert _next_rec.info.startswith('141.48.10.202@2023-01-17_15:55:49')
    _info_token = get_enclosed(_next_rec.info)
    _last_info = json.loads(_info_token)['info']
    assert 'processing https://opendata.uni-halle.de/retrieve/b7f7f81d-e65f-4c7d-95c6-7384b184c6a9/00001051.jpg' in _last_info


def test_record_handler_quotation_fixed_ast(tmp_path):
    """
    Fix behavior when encountering data which
    has been fixed manually resulting in a mixture
    of double and single quotes

    originates from ODEM project csv list
    """

    # arrange
    path_list_sample = os.path.join(TEST_RES, 'oai-records-opendata-vd18-sample-fixed.csv')
    path_oai_list1 = tmp_path / 'oai-list.csv'
    shutil.copy(path_list_sample, path_oai_list1)
    handler = OAIRecordHandler(path_oai_list1, transform_func=transform_to_record)

    # act
    _next_rec = handler.next_record(state='ocr_fail')    
    assert _next_rec.info.startswith('141.48.10.202@2023-01-17_15:55:49')
    _info_token = get_enclosed(_next_rec.info)
    _last_info = ast.literal_eval(_info_token)['info']
    assert 'processing https://opendata.uni-halle.de/retrieve/b7f7f81d-e65f-4c7d-95c6-7384b184c6a9/00001051.jpg' in _last_info


def test_record_handler_quotation_mixture_ast(tmp_path):
    """
    Fix behavior when encountering data which
    has been fixed manually resulting in a mixture
    of double and single quotes that shall be parsed with
    standard python eval from ast module which should
    work if the key "info" is double-quotes as well
    as the total information string which might contain
    single-quotes

    originates from ODEM project csv list
    """

    # arrange
    path_list_sample = os.path.join(TEST_RES, 'oai-records-opendata-vd18-sample-mixture.csv')
    path_oai_list1 = tmp_path / 'oai-list.csv'
    shutil.copy(path_list_sample, path_oai_list1)
    handler = OAIRecordHandler(path_oai_list1, transform_func=transform_to_record)

    # act
    _next_rec = handler.next_record(state='ocr_fail')    
    assert _next_rec.info.startswith('141.48.10.202@2023-01-17_15:55:49')
    _info_token = get_enclosed(_next_rec.info)
    _last_info = ast.literal_eval(_info_token)['info']
    assert "processing 'https://opendata.uni-halle.de/retrieve/b7f7f81d-e65f-4c7d-95c6-7384b184c6a9/00001051.jpg'" in _last_info


@pytest.mark.skipif(sys.version_info >= (3,10), reason="Specific to Python <3.10")
def test_record_handler_quotation_mixture_ast_639763(tmp_path):
    """
    Behavior when encountering data like this:

    originates from ODEM project csv list

    Please note: changed from python 3.8 to 3.10
        therefore another testcase
    """

    # arrange
    path_list_sample = os.path.join(TEST_RES, 'oai-records-opendata-vd18-sample-mixture.csv')
    path_oai_list1 = tmp_path / 'oai-list.csv'
    shutil.copy(path_list_sample, path_oai_list1)
    handler = OAIRecordHandler(path_oai_list1, transform_func=transform_to_record)

    # act
    _next_rec = handler.next_record(state='ocr_busy')    
    _info_token = get_enclosed(_next_rec.info)
    with pytest.raises(SyntaxError) as _decode_err:
        ast.literal_eval(_info_token)
    assert 'invalid syntax' in _decode_err.value.args[0]


@pytest.mark.skipif(sys.version_info < (3,10), 
                    reason="Specific to Python >= 3.10")
def test_record_handler_quotation_mixture_ast_639763_python_310(tmp_path):
    """
    Behavior when encountering data like this:

    originates from ODEM project csv list

    Please note: changed from python 3.8 to 3.10
        therefore another testcase
    """

    # arrange
    path_list_sample = os.path.join(TEST_RES, 'oai-records-opendata-vd18-sample-mixture.csv')
    path_oai_list1 = tmp_path / 'oai-list.csv'
    shutil.copy(path_list_sample, path_oai_list1)
    handler = OAIRecordHandler(path_oai_list1, transform_func=transform_to_record)

    # act
    _next_rec = handler.next_record(state='ocr_busy')    
    _info_token = get_enclosed(_next_rec.info)
    with pytest.raises(SyntaxError) as _decode_err:
        ast.literal_eval(_info_token)
    assert "unmatched '}'" in _decode_err.value.args[0]


@mock.patch('requests.get')
def test_oai_load_exception_for_server_error(mock_504, tmp_path):
    """Ensure OAILoadException for Response status_code
    which indicates internal Server-Errors gets properly
    propagated upstream to caller.
    
    Because we're testing the very response status_code,
    any subsequent steps that usually require additional
    information for parsing the response content (like
    prime id) are not needed and thereforse just set
    with dummy identifier 'foo'
    """

    # arrange
    # arrange
    _req = mock_response(status_code=504)
    mock_504.return_value = _req
    record = OAIRecord('foo')
    _id = record.local_identifier
    local_dir = tmp_path / "WORKDIR" / _id
    store_dir = tmp_path / "STORE" / "dd" / _id
    local_dir.mkdir(parents=True)
    store_dir.mkdir(parents=True)
    key_images = 'MAX'
    local_dst = str(local_dir) + '/' + _id + '.xml'
    request_kwargs = dict(headers={'User-Agent': 'Smith'})

    # act
    loader = OAILoader(local_dir, base_url=OAI_BASE_URL_OPENDATA,
                       group_images=key_images,
                       post_oai=post_oai_extract_mets,
                       request_kwargs=request_kwargs)
    loader.store = LocalStore(store_dir, local_dir)

    # act
    with pytest.raises(Exception) as exc:
        loader.load(record.identifier, local_dst, 'foo')

    # assert
    assert exc.typename == 'OAILoadServerError'
    _msg = exc.value.args[0]
    assert _msg == "url 'opendata.uni-halle.de/oai/dd?verb=GetRecord&metadataPrefix=mets&identifier=foo' returned '504'"
